/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : AirScheduleSolu2.cc                                      */
/*                                                                       */
/*       Date : November 3, 1997                                         */
/* Updated at : August  22, 1998                                         */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#include <iomanip.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <fstream.h>
#include <time.h>

#include <deffunc.h>
#include <tSet.h>
#include <DcSortedSet.h>
#include <StatisticSet.h>
#include <gVector.h>
#include <List.h>
#include <BitArray.h>
#include <RandomGenerator.h>

#include "AirScheduleProblem.h"
#include "AirScheduleSolu2.h"
#include "EdgeGraph.h"
#include "Edge.h"
#include "Bin.h"
#include "Pair.h"
#include "CandidateELSet.h"
#include "PathBuilder.h"

#include "debugs.h"

/************************************************************************/
/** */
/************************************************************************/

AirScheduleSolu2::AirScheduleSolu2(AirScheduleProblem* _Problem ) 
{  
   if(DEBUG5) cout << " in AirScheduleSolu2::AirScheduleSolu2() " << endl;
   myProblem = _Problem;
   cost      = 0;
   calculated = 0;

   edgeGraph = NULL;
   debug = 1;
   candidates = NULL;
   myReset();
   if(DEBUG5) cout << " before new Cover() " << endl;
   myCover = new Cover(myProblem->getNumberOfFlights());
   calculatedCover=1;


}// end of constructor

/************************************************************************/
/*  */
/************************************************************************/

AirScheduleSolu2::AirScheduleSolu2(AirScheduleProblem* _Problem, 
				   EdgeGraph *_edgeGraph) 
{  
   if(DEBUG5) cout << " in AirScheduleSolu2::AirScheduleSolu2(pr,eg) " << endl;
   myProblem = _Problem;
   cost      = 0;
   calculated = 0;

   edgeGraph = _edgeGraph;
   debug = 1;
   candidates = NULL;
   myReset();
   if(DEBUG5) cout << " before new Cover() " << endl;

   myCover = new Cover(myProblem->getNumberOfFlights());
   calculatedCover=1;

}// end of constructor

/************************************************************************/
/*  */
/************************************************************************/

AirScheduleSolu2::AirScheduleSolu2(const AirScheduleSolu2& RHS) 
{  
   copyFrom(RHS);
   myReset();

}// end of COPY constructor

/************************************************************************/
/*  */
/************************************************************************/

AirScheduleSolu2::AirScheduleSolu2() 
{ 
   myProblem = NULL;
   cost      = 0;
   calculated = 0;
   edgeGraph  = NULL;
   candidates = NULL;
   myCover = NULL; //new Cover(1);
   calculatedCover=0;
   myReset();

}// end of Default constructor

/************************************************************************/
/*  */
/************************************************************************/

void AirScheduleSolu2::checkMem(char* mess)
{
   system("ps -l -u htozdemi | grep solveGA > ou");
   ofstream out("mem.trace",ios::app);
   ifstream rin("ou");
   char line[600];
   rin.getline(line,599);
   rin.close();
   if(line!=NULL) out << mess << endl << line << endl;
   out.close();

}// end of checkMem()

/************************************************************************/
/*  */
/************************************************************************/

void
AirScheduleSolu2::copyFrom(const AirScheduleSolu2& RHS) 
{  
   candidates = NULL;
   int i, j, k;
   ListOfEdgeLists *tmpL;
   EdgeList *tmpE;

   myProblem = RHS.myProblem;
   edgeGraph = RHS.edgeGraph;
   cost = 0;
   calculated =0;

   myEdgeLists.reset();
   myCover = new Cover(myProblem->getNumberOfFlights());
   calculatedCover=1;

   cout << " getNumOfRotations " << RHS.getNumOfRotations() << endl;

   // Make a copy of EdgeList elements
   for(i=0;i<RHS.getNumOfRotations();i++)
   {
      tmpL = RHS.myEdgeLists.getNthElement(i);

      if(tmpL!=NULL)
      { 

         for(j=0;j<tmpL->getNumberOfEdgeLists();j++)
	 {
            cout << j << "/" << tmpL->getNumberOfEdgeLists() << endl;
            tmpE = tmpL->getNthEdgeList(j);
	    if(tmpE!=NULL) insertEdgeList(*tmpE);
	    
	 }// end of for

      }// end of if 

      cout << i<< "/" << RHS.getNumOfRotations() << endl;

   }// end of for

   myReset();

}// end of copyFrom() 

/************************************************************************/
/* // */
/************************************************************************/

void
AirScheduleSolu2::dispose()
{  
   myReset();

}// end of dispose()

/************************************************************************/
/* // */
/************************************************************************/

AirScheduleSolu2::~AirScheduleSolu2() 
{  
    dispose();

}// end of destructor()

/************************************************************************/
/** Returns the problem to which this solution belongs.                 */
/************************************************************************/
AirScheduleProblem* 
AirScheduleSolu2::getProblem() const { return myProblem; }

/************************************************************************/
/** Returns the edgegraph for this solution's problem.                  */
/************************************************************************/
EdgeGraph* 
AirScheduleSolu2::getEdgeGraph() const { return edgeGraph; }

/************************************************************************/
/** Returns all the edges in the solution                               */
/************************************************************************/
EdgeList 
AirScheduleSolu2::getAllEdges()
{ 
   int i, j, k;
   ListOfEdgeLists *tmpL;
   EdgeList        *eLst;
   Edge            *edge;
   EdgeList ret(myProblem->getNumberOfFlights());

   // For each ListOfEdgeLists
   for(i=0;i<getNumOfRotations();i++)
   {  
      tmpL = getListOfEdgeListsIndex(i);
      if(tmpL!=NULL)
      { 
	 // For each EdgeList
         for(j=0;j<tmpL->getNumberOfEdgeLists();j++)
	 { 
	    eLst = tmpL->getNthEdgeList(j);
	    if(eLst!=NULL)
	    {
	       // For each Edge
               for(k=0;k<eLst->getNumberOfEdges();k++)
	       {
	          edge = eLst->getNthEdge(k);

		  if(edge!=NULL) ret.addElement(edge);

	       }// end of for

            }// end of if 

	 }// end of for

      }// end of if 

   }// end of for

   return ret; 

}// end of getAllEdges()

/************************************************************************/
/**
   Under Utilization Ratio is the ratio of non-flying times to the whole
   duty period. Of course, we want to minimize this value as much as
   possible to get the most of it from the schedule.
*/
/************************************************************************/
double 
AirScheduleSolu2::getUnderUtilizationRatio() 
{

   //
   //  non used time over the total time span will give us the ratio of 
   //  unused time in the total time spanfor the schedule at hand.       
   //

   if(!calculatedStats) CalculateStatistics();

   return (tofSpanningTimes-tofFlyingTimes)/tofSpanningTimes;

}// end of getUnderUtilizationRatio()

/************************************************************************/
/************************************************************************/
double
AirScheduleSolu2::getSquareSumOfRotationFlyingTimes(float parameter) 
{
   double ret = 0;
   double t1;
   int t, Tmin, i;
   float param = maximum(parameter,1.0);
   int indx = getNumOfRotations();
   ListOfEdgeLists *tmp;

   if(indx>0)
   {
      /*Tmin = myEdgeLists[0].getTotalFlyingTime();
      for(i=0;i<index;i++)
      {
         t = myEdgeLists[i].getTotalFlyingTime() ;
         if(Tmin>t) Tmin = t;
      }*/
   

      for(i=0;i<indx;i++)
      {
         //t = ( myEdgeLists[i].getTotalFlyingTime() - Tmin );
         //t1 = ( myEdgeLists[i].getTotalFlyingTime() / 100.0 );
	 tmp = getListOfEdgeListsIndex(i);
	 if(tmp!=NULL) 
         {
	    t1 = ( tmp->getTotalFlyingTime() / param );
            ret += (t1*t1);
         }// end of if

      }// end of for

   }// end of if

   return ret;

}// end of getSquareSumOfRotationFlyingTimes()

/************************************************************************/
/*
//
// Returns the total cost of this schedule.
//
*/
/************************************************************************/

double 
AirScheduleSolu2::getCost() 
{  double tot=0;
   unsigned i;
   float ave=0;
   float CostPerMinute = myProblem->getCostPerMinute();
   float CostPerOvernight = myProblem->getCostPerOvernight();
   float LoadBalancingFactor = myProblem->getLoadBalancingFactor();
   float dif; 
   int indx = getNumOfRotations();
   ListOfEdgeLists *tmp;

   if(!calculated)
   {  
      if(indx>0)
      {
	 /** Compute the average number of flights */
	 for(i=0;i<indx;i++) 
	 { 
	    tmp = getListOfEdgeListsIndex(i);
	    if(tmp!=NULL) ave += tmp->getTotalFlyingTime();
	 }// end of for
	 ave /= indx;

	 /** Punish the Lists which are beyond average */
	 for(i=0;i<indx;i++)
	 {
	   tmp = getListOfEdgeListsIndex(i);

           if(tmp!=NULL) 
	   { 
	      dif = tmp->getTotalFlyingTime() - ave;
	      if(dif<0) dif = -1.0 * dif;
	      tot += ((dif>0.1*ave? LoadBalancingFactor:1.0)*
		       tmp->getCost(CostPerMinute,CostPerOvernight));
           }// end of if

         }// end of for

         cost = tot;
         calculated = (char)1;

      }// end of if

   }// end of if

   return cost;

}// end of getCost()

/************************************************************************/
/*
//
// Returns the real total cost of this schedule.
//
*/
/************************************************************************/

double 
AirScheduleSolu2::getCost2() 
{  double tot=0;
   unsigned i;
   float ave=0;
   float CostPerMinute       = myProblem->getCostPerMinute();
   float CostPerOvernight    = myProblem->getCostPerOvernight();
   float LoadBalancingFactor = myProblem->getLoadBalancingFactor();
   float dif; 
   int indx = getNumOfRotations();
   ListOfEdgeLists *tmp;

   if(indx>0)
   {
      /** Compute the average number of flights */
      for(i=0;i<indx;i++) 
      { 
         tmp = getListOfEdgeListsIndex(i);
         if(tmp!=NULL) ave += tmp->getTotalFlyingTime();
      }// end of for
      ave /= indx;

      /** Punish the Lists which are beyond average */
      for(i=0;i<indx;i++)
      {
	  tmp = getListOfEdgeListsIndex(i);

          if(tmp!=NULL) 
	  { 
	      dif = tmp->getTotalFlyingTime() - ave;
	      if(dif<0) dif = -1.0 * dif;
	      tot += ((dif>0.1*ave? LoadBalancingFactor:1.0)*
		       tmp->getCost(CostPerMinute,CostPerOvernight));

          }// end of if

       }// end of for

   }// end of if

   return tot;

}// end of getCost2()

/************************************************************************/
/*  Returns the real total cost of this schedule.  */
/************************************************************************/
double AirScheduleSolu2::getDCost() 
{  
   double tot=0;
   unsigned i;
   int indx = getNumOfRotations();
   ListOfEdgeLists *tmp;

   for(i=0;i<indx;i++) 
   { 
      tmp = getListOfEdgeListsIndex(i);
      if(tmp!=NULL) tot += tmp->getDCost(myProblem);

   }// end of for

   //cout << " tot " << tot << endl;
   return tot;

}// end of getDCost()

/************************************************************************/
/************************************************************************/
void AirScheduleSolu2::myReset( )
{
  calculated      = 0;
  calculatedStats = 0;
  calculatedUUR   = 0;
  minRotationTime = maxRotationTime = aveRotationTime = 0;
  stdRotationTime = 0;
  notEndsWStarted = notFromBC = numOfDeadheadings = 0;
  tofSpanningTimes = tofFlyingTimes = 0;

}// end of myReset()

/************************************************************************/
/************************************************************************/
void AirScheduleSolu2::CalculateStatistics( )
{  
   double ret  = 0;
   double diff = 0;
   double    a = 0;
   unsigned i, k1, k2;
   float ave=0;
   int indx = getNumOfRotations();
   ListOfEdgeLists *tmp;
   double    minR = FLT_MAX;
   double    maxR = FLT_MIN;
   double    fTime;
   int       NumOfFlights = myProblem->getNumberOfFlights();
   int       *deadheadings = new int[NumOfFlights];
   Edge      *tedge;
   EdgeList  *el;
   int       totDH=0;
   double    totalSpan = 0;
   double    totalFlying = 0;

   for(i=0;i<NumOfFlights;i++) deadheadings[i]=0;

   int notEndsWhereItStarted = 0;
   int notStartFromBaseCity  = 0;

   if(indx>0)
   {
      /** Compute the average number of flights */
      for(i=0;i<indx;i++) 
      { 
         tmp = getListOfEdgeListsIndex(i);
         if(tmp!=NULL) 
	 { 
	    ave += tmp->getTotalFlyingTime();
	    totalSpan += tmp->getTotalTimeSpan();

	 }// end of if

      }// end of for
      totalFlying = ave;
      ave /= indx;

      /** Compute the average number of flights */
      for(i=0;i<indx;i++) 
      { 
         tmp = getListOfEdgeListsIndex(i);

         if(tmp!=NULL) 
	 { 
	    fTime = tmp->getTotalFlyingTime();
	    a = ave-fTime;
	    diff += (a*a);

	    if( fTime>maxR ) maxR = fTime;
	    else if( fTime<minR ) minR = fTime;

            if( !tmp->doesStartFromBaseCity() ) notStartFromBaseCity++;

            if( tmp->getStartCityIndex() != tmp->getDestinationCityIndex() )
		notEndsWhereItStarted++;


            for(k1=0;k1<tmp->getNumberOfEdgeLists();k1++)
	    {
               el = tmp->getNthEdgeList(k1);

               if( el!=NULL )
               { 
		  tedge = NULL;
                  for(k2=0;k2<el->getNumberOfEdges();k2++)
                  {
                     tedge = el->getNthEdge(k2);
                     if( tedge!=NULL ) deadheadings[tedge->getStartNode()]++;

                  }// end of for

                  if( tedge!=NULL ) deadheadings[tedge->getEndNode()]++;

               }// end of if

            }// end of for

	 }// end of if

      }// end of for

      ret = sqrt(diff/indx);

   }// end of if

   for(i=0;i<NumOfFlights;i++) 
       if(deadheadings[i]>0) totDH += (deadheadings[i]-1);

   minRotationTime   = minR;
   maxRotationTime   = maxR;
   aveRotationTime   = ave;
   stdRotationTime   = ret;
   numOfDeadheadings = totDH;
   notEndsWStarted   = notEndsWhereItStarted ;
   notFromBC         = notStartFromBaseCity ;
   tofSpanningTimes  = totalSpan ;
   tofFlyingTimes    = totalFlying ;
   calculatedStats   = 1;

   delete[] deadheadings;

}// end of getRotationSTD()

/************************************************************************/
/************************************************************************/

double 
AirScheduleSolu2::getRotationSTD( double& min, double& max, 
				  double& average, int& notEndsWhereItStarted,
				  int& notStartFromBaseCity, int& deadheads) 
{  
   if(!calculatedStats) CalculateStatistics();

   min = minRotationTime ;
   max = maxRotationTime ;
   average = aveRotationTime ;
   double ret = stdRotationTime ;
   deadheads = numOfDeadheadings ;
   notEndsWhereItStarted = notEndsWStarted ;
   notStartFromBaseCity = notFromBC ;

   return ret;

}// end of getRotationSTD()

/************************************************************************/
/*
//
// Returns the number of base cities.
//
*/
/************************************************************************/

AirScheduleSolu2::getNumberOfBaseCities() 
{
   unsigned i;
   int indx = getNumOfRotations();
   tSet<int> CityIndexSet;
   ListOfEdgeLists *tmp;
   int ret, cityIndex;
   Edge *edge;
   EdgeList *first;

   for(i=0;i<indx;i++)
   {
      /*************************************/	
      /**get the index of the first flight */
      /*************************************/	

      tmp  = getListOfEdgeListsIndex(i);

      if(tmp!=NULL)
      {
	 first = tmp->getFirstEdgeList();

	 if(first!=NULL)
	 {
	    edge = first->getFirstEdge();

	    if(edge!=NULL) CityIndexSet.addElement(edge->getStartCityIndex());

	 }// end of if

      }// end of if

   }// end of for
  
   return CityIndexSet.getSize();

}// end of getNumberOfBaseCities()

/************************************************************************/
/*
//
// sort the columns according to the number of flights that can cover
//
*/
/************************************************************************/

void 
AirScheduleSolu2::sortDes()
{  unsigned i, j;
   Column Temp;
   int indx = getNumOfRotations();

   /***********************************************************************/
   /** sort the columns according to the number of flights that can cover */
   /***********************************************************************/

   cout << " index " << indx << endl;

   /*
   for(i=0;i<index-1;i++)
     for(j=i+1;j<index;j++)
        if(myEdgeLists[i].getNumberOfEdgeLists()<myEdgeLists[j].getNumberOfEdgeLists())
	   swap(myEdgeLists[i],myEdgeLists[j]);
   */

   cout << " I am out of sortDes() " << endl;

}// end of sortDes()

/************************************************************************/
/**
    + For each uncovered flight

      - looak at all the bins(knapsacks) to see whether this flight
	can be put into. If there is such bin, pick the one with minimum
	cost value.

      - find the flight which ends at Cj with minimum cost.

      - Prefer the flight which comes from bin based comparison. If
	that's so, then add this edge to the corresponding bin.

      - If the flight comes from the second choice, then define another
        bin and place this edge to the new bin.

    + Define an edge list from each bin and place them to the solution.

    + applyRotationCheck: checks the validity of each rotation.

*/
/************************************************************************/

void 
AirScheduleSolu2::solveGreedyBin()
{
   unsigned NumOfFlights = myProblem->getNumberOfFlights();
   char *covered = new char[NumOfFlights];

   int i, j, minFound, binMinFound, binMinI, minI;
   float minCost, binMinCost;
   EdgeGraph currEG = *edgeGraph;
   EdgeList* el;
   Edge* edge;
   Bin* myBins = NULL;
   int numberOfBins = 0;

   discardAllListOfEdgeLists();

   for(j=0;j<NumOfFlights;j++) covered[j]=(char)0;
   if(DEBUG2) printCover(NumOfFlights,covered);

   for(j=NumOfFlights-1;j>=0;j--)
   {
      
      if(covered[j]==0)
      {
         /** For each current node of bin find minimum */
         binMinFound = 0; binMinCost = FLT_MAX;
         for(i=0;i<numberOfBins;i++) 
	   if(currEG.doesEdgeExist(j,myBins[i].getCurrentNode()) && 
	      currEG.getCost(j,myBins[i].getCurrentNode())<binMinCost)
	   {
	      binMinI = i;
	      binMinFound = 1;
	      binMinCost = currEG.getCost(j,myBins[i].getCurrentNode());
	   }

         minFound = 0; 
         minCost = FLT_MAX;
         for(i=j-1;i>=0;i--) 
	    if(currEG.doesEdgeExist(i,j) && 
	       currEG.getCost(i,j)<minCost&&covered[i]==0)
	    {
	       minCost = currEG.getCost(i,j);
	       minFound = 1;
	       minI = i;
	    } 

         //if(binMinFound == 1 && binMinCost < minCost)
         if(binMinFound == 1 )//&& binMinCost < minCost)
	 {
	    if(DEBUG2) cout << " Minimum Found! " << endl;

	    edge = edgeGraph->getEdgeFromTo(j,myBins[binMinI].getCurrentNode());
	    myBins[binMinI].putEdge(edge);
	    covered[j]=(char)1;
	    minFound = 0;

	 }// end of if

         if(minFound == 1)
	 {
	    if(DEBUG2) cout << " Minimum Found! " << endl;
	    edge = edgeGraph->getEdgeFromTo(minI,j);

	    /** Define a new bin */
	    Bin* Temp = new Bin[numberOfBins+1];
	    for(i=0;i<numberOfBins;i++) Temp[i] = myBins[i];
            Temp[i].putEdge(edge);
	    if(numberOfBins>0) delete[] myBins;
	    numberOfBins++;
	    myBins = Temp;

	    covered[j]=(char)1;
	    covered[minI]=(char)1;

	 }// end of if

      }// end of if

      if(DEBUG2) cout << " j: " << j << endl;

   }// end of for
    

   if(DEBUG2) cout << " ****** Before PrintCover " << endl;
   if(DEBUG2) printCover(NumOfFlights,covered);

   if(DEBUG2) cout << " ****** Before insertEdgeList " << endl;
   for(i=0;i<numberOfBins;i++) insertEdgeList(myBins[i].getEdgeList());
   if(DEBUG2) cout << " Before applyRotationCheck " << endl;
   if(DEBUG2) cout <<  *this << endl;
   if(DEBUG2) writeStat(cout);

   char fileName[30];
   sprintf(fileName,"%s","solveGreedyBin.out");
   ofstream oStat(fileName);
   oStat  << " Before applyRotationCheck " << endl;
   oStat <<  *this << endl;
   writeStat(oStat);

   discardAllListOfEdgeLists();
   applyRotationCheck(NumOfFlights, myBins, numberOfBins);
   if(DEBUG2) cout << " After applyRotationCheck " << endl;
   if(DEBUG2) cout <<  *this << endl;
   if(DEBUG2) writeStat(cout);
   oStat << " After applyRotationCheck " << endl;
   oStat <<  *this << endl;
   writeStat(oStat);
   oStat.close();

   if(numberOfBins) delete[] myBins;
   delete[] covered;

}// end of solveGreedyBin()

/************************************************************************/
/** 
    + For each uncovered flight(backwards)

      - Let Cj represents the uncovered flight at hand.

      - Find a flight Ci which can connect to flight Cj with minimum
        cost. If such Ci exist, then

	  * define a new EdgeList

	  * add edge(Ci->Cj) to this list

	  * add this edge list to the solution

	  * update cover for Ci and Cj

	  * Cj=Ci

*/
/************************************************************************/

void 
AirScheduleSolu2::solveGreedy()
{
   unsigned NumOfFlights = myProblem->getNumberOfFlights();
   char *covered = new char[NumOfFlights];

   int i, j, minFound, minI;
   float minCost;
   EdgeGraph currEG = *edgeGraph;
   EdgeList* el;
   Edge* edge;
   Edge* edge2;

   discardAllListOfEdgeLists();

   for(j=0;j<NumOfFlights;j++) covered[j]=(char)0;
   if(DEBUG2) printCover(NumOfFlights,covered);

   for(j=NumOfFlights-1;j>=0;j--)
   {
      
      if(covered[j]==0)
      {
	 minFound = 1; 

	 while(minFound)
	 {
	    minFound = 0; 
	    minCost = FLT_MAX;
	    for(i=j-1;i>=0;i--) 
	       if( currEG.doesEdgeExist(i,j) && currEG.getCost(i,j)<minCost &&
	           covered[i]==0)
	       {
	          minCost = currEG.getCost(i,j);
	          minFound = 1;
	          minI = i;

	       }// end of if 

            if(minFound == 1)
	    {
	       if(DEBUG2) cout << " Minimum Found! " << endl;
	       el = new EdgeList(edgeGraph->getEdgeFromTo(minI,j));

	       //el->addEdge(edgeGraph->getEdgeFromTo(minI,j));
	       insertEdgeList(*el);
	       covered[j]=(char)1;
	       covered[minI]=(char)1;

               delete el; // TIMUCIN

	    }// end of if

	 }// end of while

      }// end of if

   }// end of for
    
   if(DEBUG2)
   { 
      cout << " ***** before printCover " << endl; 
      printCover(NumOfFlights,covered);
      cout << " ***** After printCover " << endl; 

   }// end of if

   delete[] covered;

   if(DEBUG2) cout << " ***** After delete[] covered " << endl; 

}// end of solveGreedy()

/***************************************************************************/
/**
* Assumes that all the schedule is already developed and it is time to
* check the rotations validity.
*/
/***************************************************************************/
void
AirScheduleSolu2::applyRotationCheck(int NumberOfFlights, Bin *myBins, 
				     int binSize)
{
   EdgeGraph currEG = *edgeGraph;
   EdgeList *el;
   EdgeList curr;
   Edge *edge;
   char *covered = new char[NumberOfFlights];
   int i,j;

   for(i=0;i<NumberOfFlights;i++) covered[i]=0;

   //cout <<  endl << " Before applyRotationCheck() " << endl;
   //cout << " My Beans " << binSize << endl;
   //for(i=0;i<binSize;i++) 
   //    cout << " myBin[" << i << " ] : " << myBins[i].getEdgeList() << endl;


   i=0;
   while(i<binSize)
   { 
      curr = myBins[i].getEdgeList(); 

      /** Update cover, Update Graph */
      putEdgeList(&curr,&currEG,covered);

      /** Insert the edge */
      el = decideEdgeList(&curr,&currEG,edgeGraph,covered);

      if(el!=NULL) { insertEdgeList(*el); delete el; }

      i++;

   }// end of while

 
   if( findCover(NumberOfFlights,covered) &&
       HeuristicPatchUpAtMost(NumberOfFlights,edgeGraph,&currEG,covered,2,0))
            addEdgesForPatchUp(NumberOfFlights,&currEG,covered);
   
   //cout << " after applyRotationCheck() " << *this << endl;
   delete[] covered;

}// end of applyRotationCheck()

/************************************************************************/
/*
//
// produce a greedy schedule.
//
*/
/************************************************************************/

typedef List<int> List_int;

void 
AirScheduleSolu2::solveGreedy2()
{  unsigned NumOfFlights = myProblem->getNumberOfFlights();
   unsigned *value = new unsigned[NumOfFlights];
   int i = 0, j, k, found;

   if(DEBUG3) cout << " I am in solveGreedy() " << endl; 

   City *bCities = myProblem->getBaseCities();

   WeightedRandomPickerSet<Edge*> Es = 
       edgeGraph->getLeavingEdgesFromCity(myProblem->getNumberOfBaseCities(), bCities);

   EdgeGraph currEG = *edgeGraph;
   int listFound = 0;
   int mini, minj;
   float minCost; 
   gVector<List_int> GreedySolution;
   
   for(i=NumOfFlights-2;i>=0;i--)
   {
      mini = i; minj = i+1;

      /** Find the  edge which has a minimum cost */
      if(currEG.findMinOnRow(i,&minj))
      {
	 /** We need to add this */
	 listFound = 0;

	 j = 0; 
	 int last;
	 List_int *curr;

	 while(!listFound && j<GreedySolution.getSize())
	 {
	    curr = GreedySolution.getNthElement(j);

	    if(curr!=NULL)
	    {
	       /** Assuming that it is the last added */
	       last = curr->getElement(0);

	       if(last == minj ) listFound = 1;
	       else j++;

	    } else j++;

	 }// end of while

	 if(listFound) 
	 { GreedySolution.getNthElement(j)->addElement(mini);
	 }
	 else 
	 {
	    List_int *el = new List_int();
	    el->addElement(mini);
	    GreedySolution.addElement(*el);
	    delete el; // TIMUCIN
	 }

      }// end of if_minFound

   }// end of for

}// end of solveGreedy()

/**************************************************************************/
/**                                                                       */
/**  Returns all the edges in ES which starts from ( ends at ) flight
     number res.
*/
/**************************************************************************/
EdgeList 
AirScheduleSolu2::pickEdgeFromSetHaving(WeightedRandomPickerSet<Edge*> *Es, 
	    int res)
{
    int i=0;
    EdgeList ret(Es->getSize());
    Edge* edge;

    for(i=0;i<Es->getSize();i++)
    {

	if( Es->getElementAtSafe(i,edge) &&
	    ( edge->getStartNode()==res || edge->getEndNode()==res ) )
	    ret.addEdge(edge);

    }// end of for

    return ret;

}// end of pickEdgeFromSetHaving()

/**************************************************************************/
/**                                                                       */
/** We go over all the edges and we add them to a set. If (from,to) index */
/** matches any of the elements in the "res" set. If an edge matches one  */
/** element, it gets 0.5  weight and if it matches two elements then it   */
/** gets 1.0 weight. Later, we pick the edge randomly from the current    */
/** defined set.                                                          */ 
/**                                                                       */
/**************************************************************************/
Edge**
AirScheduleSolu2::pickEdgeFromCovers(WeightedRandomPickerSet<Edge*> *Es,
                                     tSet<int> *res)
{
   
   int i;
   WeightedRandomPickerSet<int> currES(Es->getSize());
   int* ret;
   Edge** edge;
   Edge* edge1;
   int notExistFrom, notExistTo, j;

   if(DEBUG3) cout << " in pickEdgeFromCovers() " << endl;

   /*
   if(DEBUB2)
   {
      cout << " There is an Extinct node " << endl << "    { ";
      for(i=0;i<res->getSize()-1;i++) 
          if(res->getElementAtSafe(i,edge1)) cout << setw(3) << *edge1 << ", ";
      if(res->getElementAtSafe(i,edge1)) cout << setw(3) << *edge1 << " }";
   }// end of if
   */

   int a;
   i=0;
   while( i<Es->getSize() )
   {

      notExistFrom = notExistTo = j=0;

      if(Es->getElementAtSafe(i,edge1))
        while(j<res->getSize() && ( !notExistFrom || !notExistTo ) )
        {
	   if(res->getElementAtSafe(j,a))
	   {
              if( edge1->getStartNode() == a ) notExistFrom=1;
              if( edge1->getEndNode() == a ) notExistTo=1;

	   }// end of if

	   j++;

        }// end of while

      if(notExistFrom && notExistTo ) currES.addElement(i,1.0);
      else if( notExistFrom || notExistTo ) currES.addElement(i,0.5);

      i++;

   }// end of while

   /*
   if(DEBUB2)
   {
      cout << " There is an Extinct node " << endl << "    { ";
      for(i=0;i<currES->getSize()-1;i++) 
          if(currES->getElementAtSafe(i,edge1)) cout << setw(3) << *edge1 << ", ";
      if(currES->getElementAtSafe(i,edge1)) cout << setw(3) << *edge1 << " }";
   }// end of if
   */

   Edge** retEdge = (currES.randomPickSafe(a)>0?Es->getElementAt(a):Es->randomPick());

   if(DEBUG3) cout << " out of pickEdgeFromCovers() " << endl;

   return retEdge;

}// end of pickEdgeFromCovers()

/**************************************************************************/
/**                                                                       */
/*
// First look at whether there is an extinct node. 
// Extinct node: Which has no incoming/outgoing edge in the
//               current edge set.
//
// If there is such a node, we try to cover them from the 
// edge list. WRONG WRONG
//
// Randomly pick an edge and see whether putiing this edge
// cause any node to go extinct or not. If so pick an edge, 
// it covers these potential extinct nodes. 
*/
/**                                                                       */
/**************************************************************************/
Edge** 
AirScheduleSolu2::semiRandomPick(int NumOfFlights, EdgeGraph *currEG, 
		  WeightedRandomPickerSet<Edge*> *Es, char * covered)
{

   int i, notAccepted, t ;
   Edge** ret;
   Edge** edge;
   tSet<int> res(NumOfFlights);

   if(DEBUG3) cout << " in semiRandomPick() " << endl;
   /***************************************/
   /** Find out not covered extinct nodes */
   /***************************************/

   for(i=0;i<NumOfFlights;i++)
     if( covered[i]==0 && 
         ( ( currEG->getNumberOfLeavingEdges(i)+ 
	     currEG->getNumberOfIncomingEdges(i) ) == 0 ) ) res.addElement(i);

   if(DEBUG3) cout << " built the edges in semiRandomPick() " << endl;

   /** */
   if(res.getSize()>0) ret = pickEdgeFromCovers(Es,&res);
   else
    {
       if(DEBUG3) cout << " There is no Extinct node " << endl;
       edge = Es->randomPick();
       notAccepted = 0;

       if(edge!=NULL && *edge!=NULL)
       for(i=0;i<NumOfFlights;i++)
       {

	  if( covered[i] == 0 && i!=(*edge)->getStartNode() &&
				 i!=(*edge)->getEndNode() )
          {

	     t = currEG->getNumberOfLeavingEdges(i) + 
		 currEG->getNumberOfIncomingEdges(i) ;

             if(currEG->doesEdgeExist(i,(*edge)->getStartNode())) t--;
             if(currEG->doesEdgeExist((*edge)->getStartNode(),i)) t--;
             if(currEG->doesEdgeExist(i,(*edge)->getEndNode())) t--;
             if(currEG->doesEdgeExist((*edge)->getEndNode(),i)) t--;

	     if(t<1) { notAccepted = 1; res.addElement(i); }

	  }// end of if 

       }// end of for

       if(notAccepted) 
       {
          if(DEBUG1) cout << " There might be an Extinct node " << endl;
	  delete edge; // TIMUCIN
	  ret = pickEdgeFromCovers(Es,&res);

       } else ret = edge;

    }

   if(DEBUG3) cout << " out of semiRandomPick() " << endl;

    return ret;
   
}// end of semiRandomPick()

/*************************************************************************/
/** Get rid of the edges which starts or ends the currently              */
/** covered nodes.                                                       */
/*************************************************************************/
void
AirScheduleSolu2::removeEdgesStartOrEndFrom(int NumOfFlights,
		  WeightedRandomPickerSet<Edge*> *Es, char * covered)
{

   int i,j;
   Edge* pedge;

   for(i=0;i<NumOfFlights;i++)
     if((int)covered[i]>0)
     {
        j=0;

        while(j<Es->getSize())
        {

           if( Es->getElementAtSafe(j,pedge) &&
             ( pedge->getStartNode()==i || pedge->getEndNode()==i ) )
           {
	         Es->removeElementAt(j);
           }
           else j++;

        }// end of while

     }// end of if

}// end of removeEdgesStartOrEndFrom()

/************************************************************************/
/** 
     Removes the effect of including edge list from edge graph and
     cover.
*/
/************************************************************************/
void
AirScheduleSolu2::putEdgesBackFromEdgeList(EdgeList *crrEdgeList, 
					   EdgeGraph *currEG, EdgeGraph *origEG,
					   char *covered)
{
   Edge *edge;
   int m, fr, to;

   /*************************/
   /** Update covered array */
   /*************************/

   if(DEBUG3) cout << " In putEdgesBackFromEdgeList() " << *crrEdgeList << endl;

   for(m=0;m<crrEdgeList->getNumberOfEdges();m++)
   {
      edge = crrEdgeList->getNthEdge(m);

      if(edge!=NULL) 
      {
	 fr = edge->getStartNode();
	 to = edge->getEndNode();

	 if(covered[fr]>0) covered[fr] = covered[fr] - 1;
	 if(covered[to]>0) covered[to] = covered[to] - 1;

      }// end of if

   }// end of f

   /**********************/
   /** Update edge graph */
   /**********************/

   for(m=0;m<crrEdgeList->getNumberOfEdges();m++)
   {
      edge = crrEdgeList->getNthEdge(m);

      if(edge!=NULL) 
      {
	 fr = edge->getStartNode();
	 to = edge->getEndNode();

	 currEG->unMarkForLeaving(fr,origEG,covered);
	 currEG->unMarkForComing(fr,origEG,covered);

	 currEG->unMarkForLeaving(to,origEG,covered);
	 currEG->unMarkForComing(to,origEG,covered);

      }// end of if

   }// end of f

   if(DEBUG3) cout << " Out of putEdgesBackFromEdgeList() " << endl;

}// end of putEdgesBackFromEdgeList()

/************************************************************************/
/** 
    Updates the edge graph and cover

     * increments cover parameter in those flights from A0,A1,...,Ak

     * removes all the edges which starts from (ends at) each Ai from 
       the edge graph.

     * returns number of new covered nodes
*/
/************************************************************************/

//void
AirScheduleSolu2::putEdgeList(EdgeList *crrEdgeList, EdgeGraph *currEG,
			      char *covered)
{
   int ret=0;
   Edge *edge;
   int m, fr, to;

   if( DEBUG1 ) cout << " In putEdgeList() " << endl;

   for(m=0;m<crrEdgeList->getNumberOfEdges();m++)
   {
      edge = crrEdgeList->getNthEdge(m);

      if(edge!=NULL) 
      {
	 fr = edge->getStartNode();
	 to = edge->getEndNode();

	 // increment the new covered nodes counter
	 if( covered[fr]==0 ) ret++;
	 if( covered[to]==0 ) ret++;

	 covered[fr] = covered[fr] + 1;
	 covered[to] = covered[to] + 1;

	 currEG->markForLeaving(fr);
	 // if(m)
	 currEG->markForComing(fr);

	 currEG->markForComing(to);
	 // if(m)

	 
      }// end of if

   }// end of f

   if(edge!=NULL) currEG->markForLeaving(to);

   if( DEBUG1 ) cout << " Out of putEdgeList() " << endl;

   return ret;

}// end of putEdgeList()

/************************************************************************/
/**
    Updates the edge probabilities based on the difficulty
    level.
**/
/************************************************************************/
void AirScheduleSolu2::updateProbabilities(WeightedRandomPickerSet<Edge*> *Es)
{
  Edge *edge;
  float pr = 0;
  int i=0;

  for(i=0;i<Es->getSize();i++)
    if(Es->getElementAtSafe(i,edge) )
    {
       pr = myProblem->getDifficulty(edge->getStartNode());
       pr += myProblem->getDifficulty(edge->getEndNode());

       Es->setProbabilityOfElementAt(i,pr);
     
    }// end of if

}// end of updateProbabilities()

/************************************************************************/
/*
//
// Adds the edges given with Es to the schedule and returns the number
// of edges being added.
//
// Note: When we enter this method currEG contains the information
//       about uncovered nodes and there is no way to go from uncovered
//       to covered node. During the method it changes the currEG but 
//       it should leave the currEG with the same consistency showing
//       that there is no edges from uncovered nodes to covered nodes.
//
// Es : keeps the edges which connects a node with the dead-end node.
//
*/
/************************************************************************/

AirScheduleSolu2::addEdgesFromSetR(int NumOfFlights, EdgeGraph *currEG, 
                  EdgeGraph *origEG, WeightedRandomPickerSet<Edge*> *Es, 
		  char * covered, int DEBUG)
{
   return addEdgesFromSetR_U(NumOfFlights, currEG, origEG, Es, covered, 0, DEBUG);

}// end of addEdgesFromSetR()

/************************************************************************/
/*
//
// Adds the edges given with Es to the schedule and returns the number
// of edges being added.
//
// Note: When we enter this method currEG contains the information
//       about uncovered nodes and there is no way to go from uncovered
//       to covered node. During the method it changes the currEG but 
//       it should leave the currEG with the same consistency showing
//       that there is no edges from uncovered nodes to covered nodes.
//
// Es : keeps the edges which connects a node with the dead-end node.
//
*/
/************************************************************************/

AirScheduleSolu2::addEdgesFromSetR_U(int NumOfFlights, EdgeGraph *currEG, 
                  EdgeGraph *origEG, WeightedRandomPickerSet<Edge*> *Es, 
		  char * covered, int UpdateSolutions, int DEBUG)
{

   float pr;
   int i, numOfAdded=0; 
   EdgeList *el, *el1 ;
   City startCity, destinationCity, currentCity;
   Edge **edge;
   Edge **pedge;
   WeightedRandomPickerSet<Edge*> Ecurr;
   //WeightedRandomPickerSet<Edge*> *Es2 = filterEdgesNotStartFromBase(Es);
   char *Curr_covered = new char[NumOfFlights];

   if(DEBUG3) cout << " in addEdgesFromSetR() " << endl;

   updateProbabilities(Es);

   if(DEBUG)
   {
      cout << " Before addEdgesFromSet() " << endl;
      printCover(NumOfFlights,covered);
      currEG->listEdges(cout);
   }

   while(Es->getSize()>0 )
   {
      for(i=0;i<NumOfFlights;i++) Curr_covered[i]=0;

      /** Print Edges */
      
      if(DEBUG)
      {
         cout << " Edges { " ;

         for(i=0;i<Es->getSize();i++) 
         {
	    edge = Es->getElementAt(i);
	    pr   = Es->getProbabilityOfElementAt(i); 

	    if(edge!=NULL) 
	    { cout << **edge << "[ "<<pr<<" ] "; 
	      delete edge;
	    }// end of if

	    if( (i+1)%2==0) cout << endl;

         }// end of for

         cout << " } " << endl;

      }// end of if

      /**************************/
      /** Randomly pick an edge */
      /**************************/
      edge = semiRandomPick(NumOfFlights,currEG,Es,covered);

      if(DEBUG2&&edge==NULL) 
	 cout << " edge is NULL "<< __FILE__ << "," << __LINE__ <<  endl;

      el   = new EdgeList(*edge);

      if(DEBUG2) cout << " defined EdgeList() " << endl;

      /******************************************************/
      /** Add the picked edge to the edgelist and remove it */
      /** from the edge set.                                */
      /******************************************************/

      Es->removeElement(*edge);

      if(DEBUG) cout << " Current Edge List : " << *el << endl; 

      Curr_covered[(*edge)->getStartNode()] = 1;
      covered[(*edge)->getStartNode()] = ((int)covered[(*edge)->getStartNode()]+1);
      Curr_covered[(*edge)->getEndNode()] = 1;
      covered[(*edge)->getEndNode()] = ((int)covered[(*edge)->getEndNode()]+1);

      /*****************************************/
      /** find the starting and current cities */
      /*****************************************/
      destinationCity = *(*edge)->getDestinationCity();
      currentCity = *(*edge)->getStartCity();

      /********************************/
      /** Make the nodes unaccessible */
      /********************************/
      currEG->markForLeaving((*edge)->getStartNode());
      currEG->markForComing((*edge)->getEndNode());
      currEG->markForLeaving((*edge)->getEndNode());

      /*******************************************************************/
      /** As long as the number of edges less than 10 and there are some */
      /** edges which points to the current node                         */
      /*******************************************************************/
      while( el->getNumberOfEdges()<10 &&
	     currEG->getNumberOfIncomingEdges((*edge)->getStartNode(),covered) )
      {
	  if(DEBUG) cout << " ask the leaving edges " << endl;

	  /********************************************/
	  /** Get the edges coming to this start node */
	  /********************************************/

	  Ecurr = currEG->getIncomingEdges((*edge)->getStartNode(),covered);

          if(DEBUG) 
	     cout << " For " << (*edge)->getEndNode() << " " << Ecurr << endl;

          pedge = semiRandomPick(NumOfFlights,currEG,&Ecurr,covered);

	  if( el->addEdgeFront(*pedge) ) // if add is OK!
	  {

	     if(DEBUG) cout << " add another edge edges " << **pedge << endl;

	     currentCity = *(*pedge)->getStartCity();
             currEG->markForComing((*pedge)->getEndNode());
             currEG->markForLeaving((*pedge)->getStartNode());


             delete edge; edge=NULL; // TIMUCIN

	     edge = pedge;
             Curr_covered[(*edge)->getStartNode()] = 1;
	     covered[(*edge)->getStartNode()] = ((int)covered[(*edge)->getStartNode()]+1);
	  }// end of if

      }// end of while

         if(DEBUG) cout << " after while " << endl;

      currEG->markForComing((*edge)->getStartNode());

         if(DEBUG) cout << " after markForComing() " << endl;

      /*
      cout << endl << " Curr_covered : ";
      for(i=0;i<NumOfFlights;i++) cout << (int)Curr_covered[i] << " ";
      */

      //printCover(NumOfFlights,covered);

      if(DEBUG) cout << " before decideEdgeList " << *el << endl;

      if(edge!=NULL) delete edge; // TIMUCIN

      el1 = decideEdgeList(el,currEG,origEG,covered);  
      delete el;  el = el1;

      if(el!=NULL) 
      {   
	 insertEdgeList(*el);
         numOfAdded += el->getNumberOfEdges();	  
	 if( UpdateSolutions ) myProblem->updatePreSolutions(el);
         delete el; // TIMUCIN

      }// end of if

      removeEdgesStartOrEndFrom(NumOfFlights,Es,covered);

      if(DEBUG) cout << " after insertEdgeList " << (*this) << endl;

   }// end of while

   delete[] Curr_covered;

   if(DEBUG)
   {
      cout << " After addEdgesFromSet() " << endl;
      printCover(NumOfFlights,covered);
      currEG->listEdges(cout);
   }

   return numOfAdded; 

}// end of addEdgesFromSetR()

/************************************************************************/
/*
//
// Adds the edges given with Es to the schedule and returns the number
// of edges being added.
//
// Note: When we enter this method currEG contains the information
//       about uncovered nodes and there is no way to go from uncovered
//       to covered node. During the method it changes the currEG but 
//       it should leave the currEG with the same consistency showing
//       that there is no edges from uncovered nodes to covered nodes.
*/
/************************************************************************/

AirScheduleSolu2::addEdgesFromSet(int NumOfFlights, EdgeGraph *currEG, 
                  EdgeGraph *origEG, WeightedRandomPickerSet<Edge*> *Es, 
		  char * covered, int DEBUG)
{
   float pr;
   int i, numOfAdded=0; 
   EdgeList *el ;
   EdgeList *el1 ;
   City startCity, destinationCity, currentCity;
   Edge **edge;
   Edge **pedge;
   WeightedRandomPickerSet<Edge*> Ecurr;
   WeightedRandomPickerSet<Edge*> *Es2 = filterEdgesNotStartFromBase(Es);
   char *Curr_covered = new char[NumOfFlights];

   if(DEBUG)
   {
      cout << " Before addEdgesFromSet() " << endl;
      printCover(NumOfFlights,covered);
      currEG->listEdges(cout);
   }

   while(Es->getSize()>0 && Es2->getSize()>0)
   {
      for(i=0;i<NumOfFlights;i++) Curr_covered[i]=0;

      /** Print Edges */
      
      if(DEBUG)
      {
         cout << " Edges { " ;
         for(i=0;i<Es->getSize();i++) 
         {
	    edge = Es->getElementAt(i);
	    pr   = Es->getProbabilityOfElementAt(i); 
	    if(edge!=NULL)  cout << **edge << "[ "<<pr<<" ] "; 
	    if( (i+1)%2==0) cout << endl;

         }// end of for
         cout << " } " << endl;
      }

      /**************************/
      /** Randomly pick an edge */
      /**************************/
      edge = semiRandomPick(NumOfFlights,currEG,Es,covered);

      el   = new EdgeList(*edge);

      if(DEBUG) cout << " defined EdgeList() " << endl;

      /******************************************************/
      /** Add the picked edge to the edgelist and remove it */
      /** from the edge set.                                */
      /******************************************************/

      numOfAdded++;
      Es->removeElement(*edge);

      if(DEBUG2) cout << " Current Edge List : " << *el << endl; 

      Curr_covered[(*edge)->getStartNode()] = 1;
      covered[(*edge)->getStartNode()] = ((int)covered[(*edge)->getStartNode()]+1);
      Curr_covered[(*edge)->getEndNode()] = 1;
      covered[(*edge)->getEndNode()] = ((int)covered[(*edge)->getEndNode()]+1);

      /*****************************************/
      /** find the starting and current cities */
      /*****************************************/
      startCity = *(*edge)->getStartCity();
      currentCity = *(*edge)->getDestinationCity();

      /********************************/
      /** Make the nodes unaccessible */
      /********************************/
      currEG->markForComing((*edge)->getStartNode());
      currEG->markForLeaving((*edge)->getStartNode());
      currEG->markForComing((*edge)->getEndNode());

      //while( currentCity!=startCity && el->getNumberOfEdges()<10 &&
      while( el->getNumberOfEdges()<10 &&
	     currEG->getNumberOfLeavingEdges((*edge)->getEndNode(),covered) )
      {
          if(DEBUG)   cout << " ask the leaving edges " << endl;

	  Ecurr = currEG->getLeavingEdges((*edge)->getEndNode(),covered);

	  if(DEBUG) cout << " For " << (*edge)->getEndNode() << " " << Ecurr << endl;

          pedge = semiRandomPick(NumOfFlights,currEG,&Ecurr,covered);

	  el->addEdge(*pedge);
	  numOfAdded++;

          if(DEBUG)  cout << " add another edge edges " << endl;

	  currentCity = *(*pedge)->getDestinationCity();
          currEG->markForLeaving((*edge)->getEndNode());
          currEG->markForComing((*pedge)->getEndNode());


	  edge = pedge;
          Curr_covered[(*edge)->getEndNode()] = 1;
	  covered[(*edge)->getEndNode()] = ((int)covered[(*edge)->getEndNode()]+1);

      }// end of while

      if(DEBUG) cout << " after while " << endl;

      currEG->markForLeaving((*edge)->getEndNode());

      if(DEBUG) cout << " after markForLeaving() " << endl;

      if(DEBUG) cout << " before decideEdgeList " << *el << endl;

      el1 = decideEdgeList(el,currEG,origEG,covered);
      delete el; el = el1;

      if(el!=NULL) { insertEdgeList(*el); delete el; }

      removeEdgesStartOrEndFrom(NumOfFlights,Es,covered);

      Es2 = filterEdgesNotStartFromBase(Es);

      if(DEBUG) cout << " after insertEdgeList " << (*this) << endl;

   }// end of while

   delete[] Curr_covered;

   if(DEBUG)
   {
      cout << " After addEdgesFromSet() " << endl;
      printCover(NumOfFlights,covered);
      currEG->listEdges(cout);
   }// end of if

   return numOfAdded; 

}// end of addEdgesFromSet()

/************************************************************************/
/*
//
// Try to cover non covered nodes by using rest of the edges on the 
// current Graph.
//
*/
/************************************************************************/

void
AirScheduleSolu2::addEdgesForPatchUp(int NumOfFlights, EdgeGraph *currEG, 
				     char * covered, 
				     float preUsageProbability)
{
   EdgeList rest = edgeGraph->getEdgesList(covered);
   int notcovered = 0;
   int rndpick, i, a;
   Edge *picked;
   Edge *nxtEdge;
   EdgeList *el ;
   EdgeList reti ;

   // try to restore from its pre calculated solutions
   
   for(i=0;i<NumOfFlights;i++) 
      if( covered[i]==0 && 
	  RandomGenerator::generateFloat(0.0,1.0)>preUsageProbability) 
      {
	 el = myProblem->selectForNode(i,covered);
	 if(el!=NULL) 
	 {  insertEdgeList(*el);
            myCover->updateCover(el,NumOfFlights,covered);
	 }// end of if

      }// end of if 

   if(findCover(NumOfFlights,covered)==0) 
   { 
      cout << endl << " Pre-Solution is USED! " << endl; 
      return;

   } else cout << endl << " Pre-Solution is *NOT ENOUGH*! Searching " << endl; 

   if(candidates!=NULL) { delete candidates; candidates=NULL; }

   if(DEBUG1)
   {
      cout << endl << " In addEdgesForPatchUp()  " << endl; 
      cout << endl <<  " rest of the Edges " << endl << rest << endl; 
   }

   patchUpWithDeadHeadings1(NumOfFlights,covered);
   if(findCover(NumOfFlights,covered)==0) return;

   patchUpWithDeadHeadings(NumOfFlights,covered);
   if(findCover(NumOfFlights,covered)==0) return;

   if(DEBUG1) cout << endl << " NotCovered = { ";
   for(i=0;i<NumOfFlights;i++) 
       if(covered[i]==0) 
       {  notcovered++;
	  if(DEBUG1) cout << i << ",";
       }
       else	
	  {
	      rest.removeEdgesStartFrom(i);
	      rest.removeEdgesEndAt(i);
	  }

   if(DEBUG1) 
      cout << " } " << endl << " rest of the Edges " << endl << rest << endl; 

   // put some non strict cycles to the schedule first
   while( notcovered && pickPath(NumOfFlights,currEG,covered,reti,1,DEBUG1) )
   {

      a = putEdgeList(&reti,currEG,covered);

      insertEdgeList(reti);

      notcovered -= a;

      if( calculatedCover ) myCover->updateCover(&reti,NumOfFlights,covered);

   }// end of while

   while( notcovered>0 && rest.getSize() )
   {
      int rndpick = rand() % rest.getSize();
      picked = rest.getNthEdge(rndpick);

      rest.removeElementAt(rndpick);

      rest.removeEdgesStartFrom(picked->getStartNode());
      rest.removeEdgesEndAt(picked->getStartNode());
      rest.removeEdgesStartFrom(picked->getEndNode());
      rest.removeEdgesEndAt(picked->getEndNode());

      EdgeList el1(picked); // = new EdgeList(NumOfFlights);

      if(covered[picked->getStartNode()]==0) notcovered--; 
      if(covered[picked->getEndNode()]==0)   notcovered--; 

      // Update Cover
      //covered[picked->getStartNode()] = (covered[picked->getStartNode()]+1);
      //covered[picked->getEndNode()] = (covered[picked->getEndNode()]+1);

      putEdgeList(&el1,currEG,covered);

      insertEdgeList(el1);

   }// end of while

   if(DEBUG1) printCover(NumOfFlights,covered);

   if(DEBUG1) cout << endl << " Out of addEdgesForPatchUp()  " << endl; 

}// end of addEdgesForPatchUp()

/************************************************************************/
/** Adds an edge to schedule as an edge list                            */
/************************************************************************/

void
AirScheduleSolu2::addEdgeToListOfEdgeList(Edge* edge, char* covered)
{

   int i;
   EdgeList *el = new EdgeList(edge);

   insertEdgeList(*el);

   i = edge->getStartNode();
   //if(covered[i]==0) covered[i]=1; 
   covered[i] = (covered[i]+1); 
   i = edge->getEndNode();
   covered[i] = (covered[i]+1); 
   //if(covered[i]==0) covered[i]=1; 

   if( calculatedCover ) myCover->updateCover(el);

   delete el; // TIMUCIN

}// end of addEdgeToListOfEdgeList() 

/************************************************************************/
/** Produces a WeightedRandomPickerSet for uncovered nodes. O(n)        */
/************************************************************************/

WeightedRandomPickerSet<int>
AirScheduleSolu2::getUncoveredNodeSet(int NumOfFlights, char* covered)
{
   int i;
   WeightedRandomPickerSet<int> ret;

   for(i=0;i<NumOfFlights;i++) if(covered[i]==0) ret.addElement(i,1.0);

   return ret;

}// end of getUncoveredNodeSet()

/************************************************************************/
/************************************************************************/

void AirScheduleSolu2::buildCandidates( int NumOfFlights, EdgeGraph *origEG, 
     char* covered, int STRICT, int DEBUG )
{
   WeightedRandomPickerSet<Edge*> edges = origEG->getAllEdges(covered);
   gVector<EdgeList> paths;
   EdgeList e1;
   int i, j, k, mxIndex=-1, maxEdges=0;
   Edge *cEdge1;
   Edge *cEdge2;
   Pair maxi;

   candidates = new CandidateELSet(STRICT,NumOfFlights,covered);

   for(i=0;i<edges.getSize();i++)
   {
      edges.getElementAtSafe(i,cEdge1);
      EdgeList eLst(cEdge1);

      for(j=i+1;j<edges.getSize();j++)
      {
	 edges.getElementAtSafe(j,cEdge2);

	 //if( eLst.getDestinationCityIndex()==cEdge2->getStartCityIndex() ) 
	 //if( eLst.getDestinationCityIndex()==cEdge2->getStartCityIndex() &&
	 //    eLst.getEndDate()<cEdge2->getEndDate() ) 
	 if( cEdge1->getEndNode()==cEdge2->getStartNode() )
         { eLst.addEdge(cEdge2); cEdge1=cEdge2; }

      }// end of for

      if(findMaximumPath(&eLst,maxi,STRICT))
      {
	 eLst.removeFromTo(maxi.getTo()+1,eLst.getSize()-1);
	 eLst.removeFromTo(0,maxi.getFrom()-1);
	 float a = ( rand() % 100 ) / 100.0 ;

	 if(   paths.getIndex() && 
             ( eLst.getSize()>maxEdges || (a>0.5&&eLst.getSize()==maxEdges) ) )
	 {
	    maxEdges = eLst.getSize();
	    mxIndex  = paths.getIndex();
	 }
	 else
	  {

	    maxEdges = eLst.getSize();
	    mxIndex  = 0;
          }

         candidates->addEdgeList(eLst);

      }// end of if

   }// end of for

}// end of buildCandidates()

/************************************************************************/
/************************************************************************/

AirScheduleSolu2::pickPath( int NumOfFlights, EdgeGraph *origEG, char* covered,
		            EdgeList& result, int STRICT, int DEBUG )
{
   int ret = 0;

   if(candidates==NULL) 
      buildCandidates( NumOfFlights, origEG, covered, STRICT, DEBUG );

   candidates->updateSet(covered,NumOfFlights);

   ret = candidates->getMax(result);

   /*
   if(paths.getIndex())
   {
      ret = 1;
      result = *paths.getElement(mxIndex);
   }// end of if
   */

   return ret;

}// end of pickPath()

/************************************************************************/
/*
   One extension to this approach to explore the path until we exhaust
   the graph so that we can return a Path.
      isThereUncoveredMatchPath()
      explorePath()
*/
/************************************************************************/

EdgeList*
AirScheduleSolu2::isThereUncoveredMatchPath(int NumOfFlights, EdgeGraph *origEG,
	      WeightedRandomPickerSet<Edge*> *potential, char* covered, int DEBUG)
{
   EdgeList reti;
   EdgeList *ret=NULL; 
   //Edge** edge=NULL;
   Edge* edge=NULL;
   Edge*  edge1;
   int j, i=0;
   int cv1, cv2;
   int currNodeId ;
   int numOfLvEdges ;
   int selectedEdge ;

   if(DEBUG)  cout << " In isThereUncoveredMatchPath() " << endl;

   if(pickPath(NumOfFlights,origEG,covered,reti,0,DEBUG))
   {
      ret = new EdgeList(reti);

   //} else edge = isThereUncoveredMatch(NumOfFlights, potential, covered, DEBUG);
   } else edge = isThereUncoveredMatch(NumOfFlights, potential, covered, DEBUG);

   if(edge!=NULL)
   {
      ret = new EdgeList(edge);

      currNodeId = edge->getEndNode();

      if( (rand()%1) > 0.5 )

      // Explore the DFS Tree Randomly
      while( edge!=NULL && 
	     (numOfLvEdges=origEG->getNumberOfLeavingEdges(currNodeId,covered))>0 )
      {
	 WeightedRandomPickerSet<Edge*> LvEdges = origEG->getLeavingEdges(currNodeId,covered);

	 if(LvEdges.randomPickSafe(edge))
	 {
	      //currNodeId = (*edge)->getEndNode();
	      //ret->addEdge(*edge);
	      currNodeId = edge->getEndNode();
	      ret->addEdge(edge);

	 }// end of if

      }// end of while

   }// end of if

   if(DEBUG)  cout << " Out of isThereUncoveredMatchPath() " << endl;

   return ret;

}// end of isThereUncoveredMatchPath()

/************************************************************************/
/**
    Returns an edge if there is an edge which connects two uncovered
    node.
*/
/************************************************************************/

Edge*
AirScheduleSolu2::isThereUncoveredMatch(int NumOfFlights,
	      WeightedRandomPickerSet<Edge*> *potential, char* covered,
	      int DEBUG)
{
   Edge* ret=NULL; 
   Edge* edge;
   int   cv1, cv2, i=0;

   while(ret==NULL && i<potential->getSize() )
   {
      if( potential->getElementAtSafe(i,edge) )
      {
	    cv1 = edge->getStartNode();
            cv2 = edge->getEndNode();
            if( cv1>=0 && cv1 < NumOfFlights && covered[cv1]==0 && 
	        cv2>=0 && cv2 < NumOfFlights && covered[cv2]==0 ) ret = edge;
            else i++;

       } else i++;

   }// end of while

   return ret;

}// end of isThereUncoveredMatch()

/************************************************************************/
/*
// 1. Find the List of EdgeList which has the edge which contains otherNode
// 
// 2. Split this List of EdgeList and remove all the edges related to
//    this node.
//
// 3. Update currEG according to the new changes.
//
*/
/************************************************************************/

void
AirScheduleSolu2::removeNodeFromListOfEdgeLists(int NumberOfFlights,
		 int otherNode, EdgeGraph * origEG, EdgeGraph * currEG, 
		 char *covered)
{
   int from, to;
   int i, j, k, m, n, indx;
   Edge* edge;
   EdgeList* crrEdgeList;
   ListOfEdgeLists temp;
   EdgeList* el, *el1;
   int stay=1;

   if(DEBUG1) 
      cout << " In removeNodeFromListOfEdgeLists() for " << otherNode << endl;

   int indx1 = getNumOfRotations();
   ListOfEdgeLists *tmp;

   for(i=0;i<indx1&&stay;i++)
   {
      /********************************/
      /** Find the List of Edge Lists */
      /********************************/

      tmp=getListOfEdgeListsIndex(i);
      if(tmp!=NULL && (j=tmp->doesContainNodeId(otherNode))>-1)
      {
	  if(DEBUG1) cout << " List of edge list " << i << " contains " << endl;
	  temp = *tmp;
	  discardListOfEdgeLists(i);

	  if(DEBUG1) cout << " after discardListOfEdgeLists("<<i<<")"<<endl;

	  /*******************************/
	  /** Add the previous edgeLists */
	  /*******************************/
	  for(m=0;m<j;m++) 
          {
	      if(DEBUG1)
	      {  cout << " I added E# " << m << " " ;
	         cout << *temp.getNthEdgeList(m) << endl;
	      }

	      insertEdgeList(*temp.getNthEdgeList(m));

          }// end of for

	  /*********************************/
	  /** Add the following  edgeLists */
	  /*********************************/
	  for(m=j+1;m<temp.getNumberOfEdgeLists();m++) 
          {
	      if(DEBUG1)
	      {
	         cout << " I added E# " << m << " " ;
	         cout << *temp.getNthEdgeList(m) << endl;
	      }

	      insertEdgeList(*temp.getNthEdgeList(m));
          }

	  /*******************************************/
	  /** Get the EdgeList which covers the node */
	  /*******************************************/

	  crrEdgeList = temp.getNthEdgeList(j);

	  if(DEBUG1) cout << " I will put the edges back " << j << endl;

	  if(crrEdgeList==NULL&&DEBUG1) cout << " NULL " << endl;

	  /*******************/
	  /** Put edges back */
	  /*******************/

	  putEdgesBackFromEdgeList(crrEdgeList,currEG,origEG,covered);

	  if(DEBUG1) cout << " I put the edges back " << endl;

          EdgeList *first = new EdgeList(crrEdgeList->getNumberOfEdges());
          EdgeList *second = new EdgeList(crrEdgeList->getNumberOfEdges());

          n = crrEdgeList->doesContainNodeId(otherNode);

	  /********************************************************/
	  /** Place back the edges which do not contain otherNode */
	  /********************************************************/
	  for(m=0;m<crrEdgeList->getNumberOfEdges();m++)
	  {
	     edge = crrEdgeList->getNthEdge(m);

	     if(edge!=NULL)
	     {
		from = edge->getStartNode();
		to   = edge->getEndNode();

		if(from!=otherNode && to!=otherNode &&
		   currEG->doesEdgeExist(from,to))
                {
		
		   if(m<n) first->addEdge(edge);
		   else    second->addEdge(edge);

		}// end of if

             }// end of if
		     
	  }// end of for

	  if(DEBUG1)
	  {  
	     cout << " EList " << *crrEdgeList << " with node " 
		  << otherNode << endl;
	     cout << " First " << *first << endl;
	     cout << " Second " << *second << endl;

	  }// end of if

          if(first->getNumberOfEdges()>0)
	  {
	     putEdgeList(first,currEG,covered);

	     el1 = decideEdgeList(first,currEG,origEG,covered);
	     if(el1!=NULL) { insertEdgeList(*el1); delete el1; }

	  }// end of if

          if(second->getNumberOfEdges()>0)
	  {
	     putEdgeList(second,currEG,covered);
	     el1 = decideEdgeList(second,currEG,origEG,covered);
	     if(el1!=NULL) { insertEdgeList(*el1); delete el1; }

	  }// end of if

          stay=0;

          delete first;  // TIMUCIN
          delete second; // TIMUCIN

      }// end of if

   }// end of for

   if(DEBUG1) cout << " out of removeNodeFromListOfEdgeLists() " << endl;

}// end of removeNodeFromListOfEdgeLists()

/************************************************************************/
/*
//
// Obviously, addEdgesForPatchUp() could not solve the problem.
// Still, there are uncovered nodes and we used up all the edges which we
// can use. Now it is time to back up a little bit and try to cover them
// again.
//
// 1. Try to find out the edges on the original graph such that
//    these edges combines uncovered nodes. I do not think that we can
//    find something but just in case.
//
// 2. Then we need drastic action. Pick an edge from the edges related
//    to uncovered node a. It will be (a,b) or (b,a). Then find the
//    edge list who contains this node. And split this edge list. This
//    might cause that another node c might be uncovered. Don not worry.
//    Put back all the edges re: a,b,c on the current Graph and try to run the 
//    program again. Good Luck.
//
// Returns number of uncovered nodes
*/
/************************************************************************/

int
AirScheduleSolu2::HeuristicPatchUpAtMost(int NumOfFlights, EdgeGraph *origEG, 
				   EdgeGraph *currEG, char * covered,
				   int numberOfTrials, int DEBUG)
{
   int ii=-1;
   //if(ii<0) return findCover(NumOfFlights,covered);

   EdgeList rest = currEG->getEdgesList();
   int notcovered = 0;
   int rndpick, i, m;
   Edge *picked;
   Edge *nxtEdge;
   EdgeList *el, *el1 ;
   WeightedRandomPickerSet<Edge*> potential;
   WeightedRandomPickerSet<Edge*> ret;
   WeightedRandomPickerSet<int> uncoveredSet;
   int *curr;
   Edge **edge = NULL;
   int otherNode;
   int attempt=0;
   int op = 1;

   //checkMem(" In HeuristicPatchUpAtMost() ");

   if(DEBUG) cout << endl <<  " In HeuristicPatchUpAtMost() " << endl; 
   //findCover(NumOfFlights,covered);
   if(DEBUG) printCover(NumOfFlights,covered);

   /***************************************/
   /** Let's find out the uncovered nodes */
   /***************************************/

   uncoveredSet = getUncoveredNodeSet(NumOfFlights,covered);

   if(DEBUG) cout << endl <<  " Uncovered Nodes " << uncoveredSet << endl; 

   /******************************/
   /** Till we covere every node */
   /******************************/

   while( uncoveredSet.getSize() > 0 && attempt<numberOfTrials)
   {
      if(DEBUG)
         cout << endl <<  " Uncovered Nodes " << uncoveredSet << endl; 

      /**************************************************/
      /** Randomly pick a node from the uncovered nodes */
      /**************************************************/

      curr = uncoveredSet.randomPick();

      /*************************************************************/
      /** Get all the edges coming/going to/from this node on the  */
      /** original graph.                                          */
      /*************************************************************/

      potential = origEG->getLeavingEdges(*curr) + 
		  origEG->getIncomingEdges(*curr) ;

      if(DEBUG)
      {
         cout << " sizeof(potential) : " << potential.getSize() << endl;
         for(m=0;m<potential.getSize();m++)
	   if(potential.getElementAtSafe(m,picked)) 
	      cout << " [" << m << "] " << *picked << endl;

      }// end of if


      /**************************************************************/
      /** If there is an edge between two uncovered node, then try  */ 
      /** this first.                                               */
      /**************************************************************/

      el = isThereUncoveredMatchPath(NumOfFlights,origEG,&potential,covered,DEBUG); 

      op = 1;

      if(el!=NULL)
      {
	 putEdgeList(el,currEG,covered);

	 el1 = decideEdgeList(el,currEG,origEG,covered);
	 delete el; el = el1;

	 if(el!=NULL) 
	 { 
	     insertEdgeList(*el);
	     op = 0;
	     delete el; // TIMUCIN

         }// end of if

      }// end of if


      if(op&&potential.getSize()>0)
      {
         /*************************************/
         /** Pick an edge random for deletion */
         /*************************************/

	 edge = potential.randomPick();

         /************************************************/
         /** find the other node which will be uncovered */
         /************************************************/

	 if((*edge)->getStartNode()!=*curr) otherNode = (*edge)->getStartNode();
	 if((*edge)->getEndNode()!=*curr) otherNode = (*edge)->getEndNode();

         /************************************************/
         /** Remove this node from the current schedule  */
         /************************************************/

         removeNodeFromListOfEdgeLists(NumOfFlights,otherNode,origEG,
					currEG,covered); 
	 
         /************************************************/
         /** Get the all edges between uncovered nodes   */
         /************************************************/

	 ret = origEG->getAllEdges(covered);

         if(DEBUG) cout << " after origEG->getAllEdges(covered) " << endl;

         if(DEBUG) cout << endl << " Heu filtered set { " << endl;

         WeightedRandomPickerSet<Edge*> filtered(ret.getSize());

	 for(m=0;m<ret.getSize();m++)
	 {
	    if(edge!=NULL) delete edge; // TIMUCIN
	    edge = ret.getElementAt(m);

	    if(edge!=NULL && covered[(*edge)->getStartNode()]==0 &&
	                     covered[(*edge)->getEndNode()]==0 )
            {
	       filtered.addElement(*edge,origEG->getMax()-(*edge)->getCost());

	       if(DEBUG)
	       {
	          cout << (**edge) << " , ";
	          if((m%2)==0) cout << endl;

	       }// end of if 

	    }// end of if

	 }// end of for

	 if(edge!=NULL) delete edge; // TIMUCIN

	 addEdgesFromSetR(NumOfFlights,currEG,origEG,&filtered,covered,DEBUG);

      }// end of if

      //findCover(NumOfFlights,covered);
      if(DEBUG) printCover(NumOfFlights,covered);

      /***************************************/
      /** Let's find out the uncovered nodes */
      /***************************************/

      uncoveredSet = getUncoveredNodeSet(NumOfFlights,covered);

      attempt++;

      delete curr; // TIMUCIN

   }// end of while

   if(DEBUG)
   {
      printCover(NumOfFlights,covered);
      cout << endl <<  " Uncovered Nodes " ;
      cout << getUncoveredNodeSet(NumOfFlights,covered) << endl; 

   }// end of if

   if(DEBUG) cout << endl <<  " Out of HeuristicPatchUp() " << endl; 

   //checkMem(" In HeuristicPatchUpAtMost() ");

   return uncoveredSet.getSize();

}// end of HeuristicPatchUp()

/************************************************************************/
/** 
    + find out the uncovered nodes

    + remove edegs coming to/going from covered edges from the EdgeGraph

    + As long as there are edges which do not have leaving edges, then
      by using this edge set apply addEdgesFromSetR. If this can not add a
      new edge list then exit from this loop. If it does, recompute the
      dead end edges.

    + If there is uncovered nodes, apply HeuristicPathcUp

    + If HeuristicPatchUp does not cover all edges, invoke
      addEdgesForPatchUp which is a more relatex version of
      HeuristicPathcUp.

    + It can be used by evolutionary operators to make sure that the
      solution at hand can be pushed to acceptable region.
*/
/************************************************************************/

void AirScheduleSolu2::patchUp(gVector<Edge*>* abondonedEdgeSet )
{
    int cnt, i, NumberOfFlights = myProblem->getNumberOfFlights();
    char *covered = new char[NumberOfFlights];
    WeightedRandomPickerSet<Edge*> Es ;
    Edge** edge;

    if(DEBUG1) cout << " In AirScheduleSolu2::patchUp() " << endl;

    if(calculatedCover)
    {
       for(i=0;i<NumberOfFlights;i++) covered[i]=myCover->iscovered(i);

    } else findCover(NumberOfFlights,covered);

    if(DEBUG1) printCover(NumberOfFlights,covered);

    EdgeGraph currEG = *edgeGraph;

    if( abondonedEdgeSet!=NULL && abondonedEdgeSet->getSize()>0 )
    {
       for(i=0;i<abondonedEdgeSet->getSize();i++)
       {
	  edge = abondonedEdgeSet->getNthElement(i);

	  if( edge!=NULL && *edge!=NULL )
	      currEG.Mark((*edge)->getStartNode(),(*edge)->getEndNode()); 

       }// end of for

    }// end of if

    if(DEBUG1) cout << endl << " uncovered nodes = { ";
    for(i=0;i<NumberOfFlights;i++)
      if(covered[i]>0)
      {
	 currEG.markForComing(i);
	 currEG.markForLeaving(i);

      } 
      else if(DEBUG1) cout << setw(3) << i <<", ";
    if(DEBUG1) cout << " } " << endl;

    cnt=1;
    Es =  currEG.getIncomingEdgesOfDeadEnds();

    while(Es.getSize()>0 && cnt>0)
    {
       //cnt = addEdgesFromSetR(NumberOfFlights,&currEG,edgeGraph,&Es,covered,DEBUG1);
       cnt = addEdgesFromSetR_U(NumberOfFlights,&currEG,edgeGraph,&Es,covered,1,DEBUG1);
       if(DEBUG1) cout << " After addEdgesFromSet " << endl; 

       if(cnt>0) Es = currEG.getIncomingEdgesOfDeadEnds() ;  

       //findCover(NumberOfFlights,covered); TIMUCIN

    }// end of while

    if(DEBUG1) cout << " before findCover() in PathcUp " << endl; 

    // Build Difficulty List
    cnt = 0;
    myProblem->resetDifficulties();
    for(i=0;i<NumberOfFlights;i++) 
	if(covered[i]==0) { cnt++; myProblem->incrementDifficulty(i); }

    //if( findCover(NumberOfFlights,covered) &&
    if( cnt &&
	//HeuristicPatchUpAtMost(NumberOfFlights,edgeGraph,&currEG,covered,3,0))
	HeuristicPatchUp(NumberOfFlights,edgeGraph,&currEG,covered,0))
	 addEdgesForPatchUp(NumberOfFlights,&currEG,covered);

    delete[] covered;

}// end of patchUp()

/************************************************************************/
/** */
/************************************************************************/

void 
AirScheduleSolu2::printCover(ostream& out, int NumOfFlights, char * covered)
const
{
   int i, cnt=0;

   if( calculatedCover ) out << *myCover << endl;
   else
   {
   if(DEBUG1) cout << " in printCover() " << endl;

   out << endl << " covered : " << endl;

   for(i=0;i<NumOfFlights;i++) 
   { 
      if((i+1)%38==0) out << setw(2) << (int)covered[i] << endl;
      else out << (int)covered[i] << " ";

      if(covered[i]>0) cnt++;

   }// end of for

   out << endl << cnt << " number of nodes are covered " << endl;

   cnt = 0;
   out << " Uncovered = { ";
   for(i=0;i<NumOfFlights;i++)
   {
       if(covered[i]==0)
       {  out << i << " , ";
          if(((cnt+1)%16)==0) out << endl << "               ";
	  cnt++;
       }// end of if

   }// end of for

   out << "} " << endl;

   if(DEBUG1) cout << " out of printCover() " << endl;
   }// end of if

}// end of printCover()

/************************************************************************/
/************************************************************************/

AirScheduleSolu2::getNumberOfUncoveredNodes() 
{
   int NumOfFlights = myProblem->getNumberOfFlights();
   int ret = 0;
   char *cover = new char[NumOfFlights];
   ret = findCover(NumOfFlights,cover);
   delete[] cover;

   return ret;

}// end of getNumberOfUncoveredNodes()

/************************************************************************/
/** Returns number of uncovered nodes and updates the cover             */
/************************************************************************/

AirScheduleSolu2::findCover(int NumOfFlights, char * covered) 
{
   int ret = 0;
   int j, k, from, to, i=0;
   EdgeList* crrEdgeList;
   Edge*    edge;
   int indx = getNumOfRotations();
   ListOfEdgeLists *tmp;
   time_t start_t;
   time_t stop_t;
   time_t dummy;

   start_t = time(&dummy);

   if(calculatedCover)
      for(i=0;i<NumOfFlights;i++) covered[i]=myCover->iscovered(i);
   else
   {
   for(i=0;i<NumOfFlights;i++) covered[i]=0;

   /********************************/
   /** Find the List of edge lists */
   /********************************/
   for(i=0;i<indx;i++)
   {
      tmp = getListOfEdgeListsIndex(i);

      /*********************************/
      /** For each  List of Edge Lists */
      /*********************************/
      if(tmp!=NULL)
      for(j=0;j<tmp->getNumberOfEdgeLists();j++)
      {
	 
         /***********************/
         /** For each Edge List */
         /***********************/
	 crrEdgeList = tmp->getNthEdgeList(j);

	 if(crrEdgeList != NULL )
	 {
            /******************/
            /** For each Edge */
            /******************/
            for(k=0;k<crrEdgeList->getNumberOfEdges();k++)
	    {
	       edge = crrEdgeList->getNthEdge(k);

	       if(edge!=NULL)
	       {  
		  from = edge->getStartNode();
		  to = edge->getEndNode();
		  covered[from] = 1+(int)covered[from];
		  covered[to] = 1+(int)covered[to];

		  if(covered[from]>2 && DEBUG5 ) 
		     cout << " weird covered["<<from<<"] = " << (int)covered[from] << endl;
		  if(covered[to]>2 && DEBUG5 ) 
		     cout << " weird covered["<<to<<"] = " << (int)covered[to] << endl;
	       }// end of if

            }// end of for

         }// end of if

      }// end of for

   }// end of for
   }// end of if

   for(i=0;i<NumOfFlights;i++) if(covered[i]==0) ret++;
   stop_t = time(&dummy);
   if(DEBUG7) cout << " findCover took " << stop_t-start_t << endl;
  
   return ret;

}// end of findCover()

/************************************************************************/
/** fiters edges that do not start from base city */
/************************************************************************/

WeightedRandomPickerSet<Edge*>*
AirScheduleSolu2::filterEdgesNotStartFromBase(WeightedRandomPickerSet<Edge*>
*Es)
{
   WeightedRandomPickerSet<Edge*> *ret = 
	   new WeightedRandomPickerSet<Edge*>(Es->getSize());
   int i=0;
   Edge *edge;

   while(i<Es->getSize())
   {

      if(Es->getElementAtSafe(i,edge) && edge->doesStartFromBaseCity())
	  ret->addElement(edge);

      i++;

      
   }// end of while

   return ret;

}// end of filterEdgesNotStartFromBase()

/************************************************************************/
/*
//
// Randomly produce a schedule.
// It takes the nodes which has no outgoing edges 
// and tries to add them to the schedule.
//
*/
/************************************************************************/

AirScheduleSolu2::solveRandomStep1()
{  
   unsigned NumOfFlights = myProblem->getNumberOfFlights();
   int i = 0, j, k, found;
   Edge *mEdge;

   if(DEBUG1) 
      cout << " I am in solveRandomStep1() " << endl; //<< *myProblem << endl;

   City *bCities = myProblem->getBaseCities();

   WeightedRandomPickerSet<Edge*> Es ;

   if(DEBUG1) 
      cout << " after getLeavingEdgesFromCity()  " << endl; 

   EdgeGraph currEG = *edgeGraph;

   char *covered = new char[NumOfFlights];

   for(i=0;i<NumOfFlights;i++) covered[i]=0;

   discardAllListOfEdgeLists();

   if(DEBUG1) 
      cout << " before while  " << endl; 

   /*
   if(DEBUG3)
   {  
      cout << endl << " { ";
      for(i=0;i<Es.getSize();i++)
      {  
	  if(Es.getElementAtSafe(i,mEdge)) cout << *mEdge << " ";
          if(i>0 && (i%5)==0) cout << endl;
      }// end of for

      cout << " } " << endl;

   }// end of if
   */

   if(DEBUG1) 
      cout << " before Es =  currEG.getIncomingEdgesOfDeadEnds(); " << endl; 

   Es =  currEG.getIncomingEdgesOfDeadEnds();

   if(DEBUG1) 
      cout << " after Es =  currEG.getIncomingEdgesOfDeadEnds(); " << endl; 

   int cnt = 1;
   int km=1;
    
   while(Es.getSize()>0 && cnt>0 && km>0)
   {
      if(DEBUG1) 
	 cout << " Before addEdgesFromSet " << endl; 

      //cnt = addEdgesFromSetR(NumOfFlights,&currEG,edgeGraph,&Es,covered);
      cnt = addEdgesFromSetR_U(NumOfFlights,&currEG,edgeGraph,&Es,covered,1);

      if(DEBUG1) 
	 cout << " I add " << cnt << " number of edges! " << endl; 

      if(cnt>0) Es = currEG.getIncomingEdgesOfDeadEnds() ;  

      removeEdgesStartOrEndFrom(NumOfFlights, &Es, covered);

      if(DEBUG1) 
	 cout << " After getIncomingEdgesOfDeadEnds() " << endl; 

      //km = findCover(NumOfFlights,covered);
      km=0;
      for(i=0;i<NumOfFlights;i++) if(covered[i]==0) km++;

      if(DEBUG1) 
	 cout << km << " number of uncovered edges left! " << endl; 


   }// end of while

   if(DEBUG1) 
      cout << " I am out of solveRandom() " << endl; //<< *myProblem << endl;

   if(DEBUG1)
      cout << " before HeuristicPatchUp " << endl;

   if(km>0 && HeuristicPatchUp(NumOfFlights,edgeGraph,&currEG,covered))
              addEdgesForPatchUp(NumOfFlights,&currEG,covered); 

   if(DEBUG1)
      cout << " after HeuristicPatchUp " << endl;


   j=0;
   for(i=0;i<NumOfFlights;i++) j += (covered[i]>0?1:0);

   if(DEBUG1)
   {  cout << endl << j << " flights are covered!  " << endl; 
      printCover(NumOfFlights,covered); 
   }// end of if 

   delete[] covered;
   delete[] bCities;

   return j;

}// end of solveRandomStep1()

/************************************************************************/
/*
//
// Randomly produce a schedule.
// It takes the nodes which has no outgoing edges 
// and tries to add them to the schedule.
//
*/
/************************************************************************/

void 
AirScheduleSolu2::solveRandom()
{  
   int attempt = 0;
   unsigned NumOfFlights = myProblem->getNumberOfFlights();

   myReset();
   if( calculatedCover ) myCover->reset();

   while(NumOfFlights!=solveRandomStep1() && ++attempt<5) 
   {
       cout << " Again " ;

   }// end of while

   cout << attempt << endl;

}// end of solveRandom()

/************************************************************************/
/*
//
// Randomly produce a schedule.
//
*/
/************************************************************************/

void 
AirScheduleSolu2::solveRandom2()
{  
   unsigned NumOfFlights = myProblem->getNumberOfFlights();
   int i = 0, j, k, found;

   if(DEBUG1) cout << " I am in solveRandom() " << endl; //<< *myProblem << endl;

   City *bCities = myProblem->getBaseCities();

   WeightedRandomPickerSet<Edge*> Es = 
       edgeGraph->getLeavingEdgesFromCity(myProblem->getNumberOfBaseCities(), bCities);

   if(DEBUG1)
      cout << " after getLeavingEdgesFromCity()  " << endl; 

   EdgeGraph currEG = *edgeGraph;

   char *covered = new char[NumOfFlights];

   for(i=0;i<NumOfFlights;i++) covered[i]=0;

   discardAllListOfEdgeLists();

   if(DEBUG1)
      cout << " before while  " << endl; 

   /*
   cout << endl << " { ";
   for(i=0;i<Es.getSize();i++)
   {  cout << **Es.getElementAt(i) << " ";
      if(i>0 && (i%5)==0) cout << endl;
   }
   cout << " } " << endl;
   */

   Es = edgeGraph->getExtinctEdges();
   addEdgesFromSet(NumOfFlights,&currEG,edgeGraph,&Es,covered);

   Es = edgeGraph->getLeavingEdgesFromCity(myProblem->getNumberOfBaseCities(), 
					   bCities);

   addEdgesFromSet(NumOfFlights,&currEG,edgeGraph,&Es,covered);

   if(DEBUG1)
      cout << " I am out of solveRandom() " << endl; //<< *myProblem << endl;

   addEdgesForPatchUp(NumOfFlights,&currEG,covered); 

   if(DEBUG1) cout << " before HeuristicPatchUp " << endl;

   HeuristicPatchUp(NumOfFlights,edgeGraph,&currEG,covered);

   if(DEBUG1) cout << " after HeuristicPatchUp " << endl;

   j=0;
   for(i=0;i<NumOfFlights;i++) j +=(int)covered[i];

   if(DEBUG1) cout << endl << j << " flights are covered!  " << endl; 

   delete[] covered;
   delete[] bCities;

}// end of solveRandom()

/************************************************************************/
/*
// Baldwin Effect
// Applies the local search to the current solution.
//
// 0) Applies the following random number of times.
//
// 1) Picks a column and tries to merge with rest of the solution. If it
//    succeed, then gets rid of this columns and inserts the new
//    merger.
*/
/************************************************************************/

void 
AirScheduleSolu2::applyLocalSearch()
{
   int numEdgeLists, j, k, i=0;
   ListOfEdgeLists *Temp ;
   EdgeList *prev;
   EdgeList *curr;
   EdgeList *el;
   EdgeList **elArray;
   Edge *edge;
   Edge *first;
   Edge *last;
   int indx = getNumOfRotations();
   ListOfEdgeLists *tmp=NULL;

   while(i<indx)
   {
      tmp = getListOfEdgeListsIndex(i);

      if(tmp!=NULL && tmp->getNumberOfEdgeLists()>1)
      {
         Temp = new ListOfEdgeLists();
	 numEdgeLists = tmp->getNumberOfEdgeLists();
         elArray = new EdgeList*[numEdgeLists];

	 for(j=0;j<numEdgeLists;j++) elArray[j] = tmp->getNthEdgeList(j);

	 j=1;
	 while(j<numEdgeLists)
	 {
	    prev = elArray[j-1];
	    curr = elArray[j];
	    first = curr->getFirstEdge();
	    last = prev->getLastEdge();

	    if(  //last->getEndNode()==first->getStartNode() &&
	        ( first->getStartDate()-last->getEndDate()) <100 &&
                edgeGraph->doesEdgeExist(last->getEndNode(),first->getStartNode()))
            {
                el = new EdgeList();
		for(k=0;k<prev->getNumberOfEdges();k++) 
		    el->addEdge(prev->getNthEdge(k));
                edge = new Edge(edgeGraph,last->getEndNode(),first->getStartNode());
		el->addEdge(edge);
		for(k=0;k<curr->getNumberOfEdges();k++) 
		    el->addEdge(curr->getNthEdge(k));
                
		elArray[j] = el;
		for(k=j;k<numEdgeLists;k++) elArray[k-1] = elArray[k];
		numEdgeLists--;

	    } else j++;

	 }// end of while

	 if(numEdgeLists<tmp->getNumberOfEdgeLists())
	 {
	    for(j=0;j<numEdgeLists;j++) Temp->addEdgeList(*elArray[j]);

	    //myEdgeLists[i] = *Temp;
	    *tmp = *Temp;

	 }// end of if


      }// end of if

      i++;

   }// end of while

}// end of applyLocalSearch()

/************************************************************************/
/*
// Baldwin Effect
// Applies the local search to the current solution.
//
// 0) Applies the following random number of times.
//
// 1) Picks a column and tries to merge with rest of the solution. If it
//    succeed, then gets rid of this columns and inserts the new
//    merger.
*/
/************************************************************************/

void 
AirScheduleSolu2::applyLocalSearch2()
{
   int numEdgeLists, j, k, i=0;
   ListOfEdgeLists *Temp ;
   EdgeList *prev;
   EdgeList *curr;
   EdgeList *el;
   EdgeList **elArray;
   Edge *edge;
   Edge *first;
   Edge *last;
   int indx = getNumOfRotations();
   ListOfEdgeLists *tmp=NULL;
   ListOfEdgeLists *tmpj=NULL;
   int notCombined=1;

   ofstream fout("local.src");

   i=0;
   while(i<getNumOfRotations()-1)
   {
      tmp = getListOfEdgeListsIndex(i);

      notCombined=1;
      if(tmp!=NULL && tmp->getNumberOfEdgeLists()>0)
      {
	 j=i+1;
	 while(j<getNumOfRotations()&&notCombined)
	 {
            tmpj = getListOfEdgeListsIndex(j);
            if(tmpj!=NULL && tmpj->getNumberOfEdgeLists()>0)
	    { 
                if(tmp->canPrecede(*tmpj,myProblem)||tmp->canFollow(*tmpj,myProblem))
		{
		   fout << *tmp << endl << *tmpj << endl;
		   notCombined=0;
		   for(k=0;k<tmpj->getNumberOfEdgeLists();k++)
		       tmp->addEdgeList(*tmpj->getNthEdgeList(k),myProblem);
		   cout << " can combine "  << i << " and " << j << endl;
		   myEdgeLists.removeElementAt(j);
		   fout << " combined to " << endl << *tmp << endl << endl;
		}    

	    }// end of if 
	    j++;

	 }// end of while

      }// end of if

      if(notCombined) i++;

   }// end of while

   fout.close();

   removeRedundancy();

   myReset();
   calculatedStats = 0;

}// end of applyLocalSearch2()

/************************************************************************/
/************************************************************************/
#include "LocalSearch.h"

void AirScheduleSolu2::applyLocalSearch3()
{
  AirScheduleSolu2* res = LocalSearch::reorganizeRotations(this);
  if(res!=NULL&&res!=this) copyFrom(*res);

}// end of applyLocalSearch3()

/************************************************************************/
/*
//  Discard the column indicated with number from the solution.
*/
/************************************************************************/

void 
AirScheduleSolu2::discardListOfEdgeLists(int edgeListNum)
{  

   ListOfEdgeLists *el = myEdgeLists.getNthElement(edgeListNum);
   if( el!=NULL && calculatedCover ) myCover->resetCover(el);
   myEdgeLists.removeElementAt(edgeListNum);
   myReset();

}// end of discardListOfEdgeLists()

/************************************************************************/
/*
//  Discard all the lists.
*/
/************************************************************************/

void 
AirScheduleSolu2::discardAllListOfEdgeLists()
{
   myEdgeLists.reset();
   if( calculatedCover ) myCover->reset();
   myReset();

}// end of discardAllListOfEdgeLists()

/************************************************************************/
/*   */
/************************************************************************/
void AirScheduleSolu2::setProblem(AirScheduleProblem* pr) 
{ 
   myProblem=pr; 
   myReset();
   myCover = new Cover(pr->getNumberOfFlights());
   calculatedCover = 1;

}// end of setProblem()

/************************************************************************/
/*  */
/************************************************************************/
void 
AirScheduleSolu2::bubbleUpColumnsAfter(unsigned indx)
{
   int j = indx;
   Column Temp;
   ListOfEdgeLists *edgeLstj;
   ListOfEdgeLists *edgeLstj_1;

   /**************************************************/
   /** buble up the new column to keep them in order */
   /**************************************************/

   while( j<getNumOfRotations() && j>0 )
   {

      edgeLstj   = getListOfEdgeListsIndex(j);
      edgeLstj_1 = getListOfEdgeListsIndex(j-1);

      if( edgeLstj!=NULL && edgeLstj_1!=NULL &&
	  edgeLstj->getNumberOfEdgeLists()>edgeLstj_1->getNumberOfEdgeLists())
          swap(*edgeLstj,*edgeLstj_1);

      j--;

   }// end of while

}// end of bubbleUpColumnsAfter()

/************************************************************************/
/*
//  Checks whether given edgeList can merge with anybody in the current
//  solution.
*/
/************************************************************************/

AirScheduleSolu2::canMerge(EdgeList edgeList)
{
   int ret = 0;

   int i=0;
   ListOfEdgeLists *tmp;
   int minIdx = -1;

   /*
   double tofFT = FLT_MAX;

   while( i<getNumOfRotations() ) {
     if( (tmp = getListOfEdgeListsIndex(i))!=NULL && 
	  tmp->canMerge(edgeList,myProblem) && 
	 (minIdx<0 || tmp->getTotalFlyingTime()<tofFT) )  
     { minIdx = i; tofFT = tmp->getTotalFlyingTime(); } 
     i++;
   }// end of while

   if( minIdx>-1 ) {
      if( (tmp = getListOfEdgeListsIndex(minIdx))!=NULL ) {
	tmp->addEdgeList(edgeList,myProblem);
	ret = 1;
	if( calculatedCover ) myCover->updateCover(&edgeList);
        myReset();
      }// end of if
   }// end of if
   */

   double tofFT = FLT_MAX;

   while( i<getNumOfRotations() ) {
     if( (tmp = getListOfEdgeListsIndex(i))!=NULL && 
	  tmp->canMerge(edgeList,myProblem) && 
	  (minIdx<0 || tmp->getUtilizationRatio()<tofFT ) ) 
     { minIdx = i; tofFT = tmp->getUtilizationRatio(); } 
     i++;
   }// end of while

   if( minIdx>-1 ) {
      if( (tmp=getListOfEdgeListsIndex(minIdx))!=NULL ) {
	tmp->addEdgeList(edgeList,myProblem);
	ret = 1;
	if( calculatedCover ) myCover->updateCover(&edgeList);
	//cout << " merged " << *tmp << endl;
        myReset();
      }// end of if
   }// end of if

   return ret;

}// end of canMerge()

/************************************************************************/
/*
//  inserts the corresponding EdgeList to the solution.
//  First tries to look at any merger with the current List of
//  edgeLists. If so, it uses this.
*/
/************************************************************************/

void 
AirScheduleSolu2::insertEdgeList(EdgeList edgeList)
{ 

   if(DEBUG3) cout << " index " << getNumOfRotations() << endl;

   if(doesCoverUncovered(myCover,&edgeList)&&!canMerge(edgeList))
   {
       if(DEBUG3) cout << " canNotMerge " << endl;

       ListOfEdgeLists nListOfEdgeList(edgeList);

       myEdgeLists.addElement(nListOfEdgeList);

       // Update Cover
       if( calculatedCover ) myCover->updateCover(&nListOfEdgeList);
       //cout << " added " << edgeList << endl;

   }// end of if()

   myReset();
  
}// end of insertEdgeList()

/************************************************************************/
/************************************************************************/

AirScheduleSolu2::doesCoverUncovered(Cover* cvr, EdgeList* edgeList)
{ 
   int retValue = 0;
   int i=0;
   EdgePtr aEdge;
   while(!retValue && i<edgeList->getNumberOfEdges()) {
     if((aEdge=edgeList->getNthEdge(i))!=NULL) 
	retValue = (!cvr->iscovered(aEdge->getStartNode())||!cvr->iscovered(aEdge->getEndNode()));
     i++;
   }// end of while

   return retValue;

}// end of doesCoverUncovered()

/************************************************************************/
/* */
/************************************************************************/

void
AirScheduleSolu2::addListOfEdgeList(ListOfEdgeLists eList)
{
   int i;
   EdgeList *el;

   for(i=0;i<eList.getNumberOfEdgeLists();i++) 
   {
      el = eList.getNthEdgeList(i);

      if(  el->doesStartFromBaseCity() && el->doesEndAtBaseCity() &&
	 ( el->getStartCityIndex()==el->getDestinationCityIndex() ) )
	  insertEdgeList(*el);
   }

}// enmd of addListOfEdgeList)

/************************************************************************/
/**  Assignment operator */
/************************************************************************/

AirScheduleSolu2 AirScheduleSolu2::operator=(const AirScheduleSolu2& RHS) 
{  

   dispose();
   copyFrom(RHS);

   return (*this);

}// end of operator=() 

/************************************************************************/
/** */
/************************************************************************/

AirScheduleSolu2::operator==(const AirScheduleSolu2& RHS)
{  
   int i, j, found = 0;
   int ret = ( myProblem == RHS.myProblem );
   ret = ret && ( getNumOfRotations() == RHS.getNumOfRotations() );
   ListOfEdgeLists *tmpi;
   ListOfEdgeLists *tmpj;
   int index = getNumOfRotations();

   for(i=0;ret && i<index ; i++)
   {
     found = 0;

     tmpi = getListOfEdgeListsIndex(i);

     if(tmpi!=NULL)
        for( j=0; !found && j<index; j++)
        {
	   tmpj = getListOfEdgeListsIndex(j);
	   if( tmpj!=NULL && *tmpi == *tmpj ) found=1;
        }// end of for

     ret = found;

   }// end of for

   cout << " they are equal " << endl;

   //if(DEBUG1) cout << (*this) << " is" << (ret?" ":" not ") 
   //                << "equal to " << RHS << endl;

   return ret;

}// end of operator==()

/************************************************************************/
/*  */
/************************************************************************/

void
AirScheduleSolu2::nicePrint(ostream &os) 
{  
   int    i, j, k1, k2, flightIndex, numOfCities;
   float  CostPerMinute    = myProblem->getCostPerMinute();
   float  CostPerOvernight = myProblem->getCostPerOvernight();
   int    NumOfFlights     = myProblem->getNumberOfFlights();
   char   *covered     = new char[NumOfFlights];
   double totalFlyTime = 0;
   ListOfEdgeLists *tmpL;
   Edge      *tedge;
   EdgeList  *el;

   //cout << " I am in operator<<(AirScheduleSolu2) " << endl;

   os << endl;

   if( calculatedCover )
      for(i=0;i<NumOfFlights;i++) covered[i] = myCover->iscovered(i);
   else findCover(NumOfFlights,covered);

   if( calculatedCover ) os << *myCover << endl;
   else printCover(os,NumOfFlights,covered);

   unsigned tot=0;

   cout << " I am in operator<<(AirScheduleSolu2) with " << getNumOfRotations() 
	<< " lists! " << endl;

   for(i=0;i<getNumOfRotations();i++)
   {  
      os << setw(3) << i << " <--> "; 

      tmpL = getListOfEdgeListsIndex(i);
      if(tmpL!=NULL)
      { 

         //cout << " Before myEdgeLists[i].nicePrint(os); " << endl; 
         tmpL->nicePrint(os); 
         tot += tmpL->getNumberOfEdges(); 

         os << " Total Flying Time : " << tmpL->getTotalFlyingTime() << endl;
         totalFlyTime +=tmpL->getTotalFlyingTime();
         os << " Costs us          : " << tmpL->getDCost(myProblem);

         //TIMUCIN os << myEdgeLists[i].getCost(CostPerMinute,CostPerOvernight);

      } else  cout << " myEdgeLists[ "<<i<<"] is empty! " ;

      os << endl;

   }// end of for

   // Total DeadHeadings
   int totDH=0;

   double ave, min, max, std;
   int    notStartFromBase = 0, notEndsWhereStarted = 0;
   std = getRotationSTD(min,max,ave,notEndsWhereStarted,notStartFromBase,totDH);

   os << " There are " << tot << " flights " ;
   os << (tot>myProblem->getNumberOfFlights()?" TOO_MANY_FLIGHTS ":" ");
   if(calculated)  os << "  cost is " << cost << endl;
   else os << "  cost is NOT_CALCULATED" << endl;

   os << "Fitness Value ...........: " 
      << getSquareSumOfRotationFlyingTimes() << endl;

   os << "Total Flying Time is are : " 
      << totalFlyTime << " minutes " << endl ;

   os << "Number of Deadheadings ..: " 
      << totDH << endl ;

   os << "Number of Required Crew .: " 
      << getNumOfRotations() << endl ;

   os << "Cost ....................: " 
      << getCost2() << endl ;

   os << "DCost ...................: " 
      << getDCost() << endl ;

   os << "Maximum Rotation Time ...: " 
      << max << endl;

   os << "Minimum Rotation Time ...: " 
      << min << endl;

   os << "Average Rotation Time ...: " 
      << ave << endl;

   os << "STD of Rotation Times ...: " 
      << std << endl;

   os << "Not from base city ......: " 
      << notStartFromBase << endl;

   os << "Not ends where it started: "
      << notEndsWhereStarted << endl;

   j=0;
   for(i=0;i<NumOfFlights;i++) if( covered[i]==0 ) j++;

   os << "Num. of Uncovered Flights: "
      //<< getNumberOfUncoveredNodes() << endl;
      << j << endl;

   delete[] covered;

}// end of nicePrint()

/************************************************************************/
/*  */
/************************************************************************/

ostream & 
operator<<(ostream& os, AirScheduleSolu2& RHS)
{  
   RHS.nicePrint(os);
   return os;

}// end of operator<<()

/************************************************************************/
/*
   numberOfListOfEdgeLists
   NumberOfEdgeList0
   NumberOfEdge0
   from0 to0 #Edge0 in first edge list
   from1 to1 #Edge1 in first edge list
   ..
   NumberOfEdge1
   from0 to0 #Edge0 in second edge list
   from1 to1 #Edge1 in second edge list
   ..
   NumberOfEdgeList1
*/
/************************************************************************/
void AirScheduleSolu2::writeCoded(ostream& os)
{
    int i, j, k;
    int numberOfLofELsts = getNumOfRotations();
    int numberOfELsts;
    int numberOfEdges;
    ListOfEdgeLists *lst;
    EdgeList *eLst;
    Edge     *edge;

    // Write number of ListOfEdgeLists
    os << numberOfLofELsts << endl;

    for(i=0;i<numberOfLofELsts;i++)
    {
       lst = getListOfEdgeListsIndex(i);

       if(lst != NULL )
       {
	  numberOfELsts = lst->getNumberOfEdgeLists();

          // Write number if EdgeLists
          os << numberOfELsts << endl;

          for(j=0;j<numberOfELsts;j++)
          {
             // Get EdgeList
	     eLst = lst->getNthEdgeList(j);

	     if(eLst!=NULL)
	     {
	        numberOfEdges = eLst->getNumberOfEdges();

                // Write number of Edges
                os << numberOfEdges << endl;

                for(k=0;k<numberOfEdges;k++)
	        {
                   // Write Edge
		   edge = eLst->getNthEdge(k);
		   if(edge!=NULL) 
		      os << edge->getStartNode() << " " 
			 << edge->getEndNode() << endl;
		   else
		      os << -1 << " " << -1 << endl;
             
	        }// end of for

	     }// end of if

          }// end of for

       }// end of if

    }// end of for

}// end of writeCoded()

/************************************************************************/
/*
   numberOfListOfEdgeLists
   NumberOfEdgeList0
   NumberOfEdge0
   from0 to0 #Edge0 in first edge list
   from1 to1 #Edge1 in first edge list
   ..
   NumberOfEdge1
   from0 to0 #Edge0 in second edge list
   from1 to1 #Edge1 in second edge list
   ..
   NumberOfEdgeList1

*/
/************************************************************************/

void AirScheduleSolu2::readCoded(istream& is)
{
    int numberOfLofELsts;
    int numberOfELsts;
    int numberOfEdges;
    int i, j, k;
    int from, to;
    EdgeList *el;

    // Read number of ListOfEdgeLists
    is >> numberOfLofELsts;

    for(i=0;i<numberOfLofELsts;i++)
    {
       cout << " before reading number of EdgeLists  " << endl;
       // Read number of EdgeLists
       is >> numberOfELsts;

       cout << "#el "<< i << "/" << numberOfLofELsts << endl; 

       for(j=0;j<numberOfELsts;j++)
       {
          // Read number of Edges
          is >> numberOfEdges;
	  cout << "#e "<< numberOfEdges << endl; 

	  // Define EdgeList
	  el = new EdgeList(numberOfEdges);

          for(k=0;k<numberOfEdges;k++)
	  {
	     is >> from >> to;
	     el->addEdge(edgeGraph->getEdgeFromTo(from,to));
             
	  }// end of for

	  // insert this EdgeList to the solution
	  cout << *el << endl;
	  if(el->isValid(myProblem)) insertEdgeList(*el);
	  else { cout << *el << " is not Valid! " << endl;
		 char aaa;
		 cin >> aaa; 
	       }
	  cout << " edge inserted " << endl;
	  delete el;
	  cout << " edge deleted " << j << endl;

       }// end of for

    }// end of for

    cout << " before removeRedundancy() " << endl;

    removeRedundancy();

}// end of readCoded()

/************************************************************************/
/**  
   writes:
    *) # of rotations
    *) miniumum flying time
    *) maximum flying time
    *) average flying time
    *) variance of flying times
    *) fitness value
    *) real cost value
*/
/************************************************************************/

void AirScheduleSolu2::writeStat(ostream& os)
{  

   //int    deadheadings = 0;
   //int    nStartFromBase = 0;
   double tot=0;
   int  numberOfUncovered = getNumberOfUncoveredNodes();
   //ListOfEdgeLists *tmp;
   int indx = getNumOfRotations();

   if(DEBUG5) 
      cout << " in AirScheduleSolu2::writeStat() of AirScheduleSolu2 "<< endl;

   os << indx << " "; 

   //float *T;
   //T=fTimes.min()      ; os << *T << " "; delete T;
   //T=fTimes.max()      ; os << *T << " "; delete T;
   //T=fTimes.average()  ; os << *T << " "; delete T;
   //T=fTimes.variance() ; os << *T << " "; delete T;
   //os << (getSquareSumOfRotationFlyingTimes() * index) << " ";
   //os << (getUnderUtilizationRatio() * index) << " ";
   //os << getCost() << " " << tot << " " << numberOfUncovered ;

   double stdofrotationtimes = 0.0;
   double minofrotationtimes = 0.0;
   double maxofrotationtimes = 0.0;
   double aveofrotationtimes = 0.0;
   int    notendswherestarted = 0;
   int    notstarsfrombase    = 0;
   int    numofdeadheads      = 0;

   stdofrotationtimes = getRotationSTD(minofrotationtimes,
			         maxofrotationtimes,aveofrotationtimes,
			         notendswherestarted,notstarsfrombase,
		                 numofdeadheads);

   os << " " << numberOfUncovered  
      << " " << notstarsfrombase 
      << " " << notendswherestarted 
      << " " << numofdeadheads 
      << " " << getDCost() ;

}// end of writeStat() 

/************************************************************************/
/**  Finds the longest sequence such as it starts from the base city    */
/**  and ends at the same base city                                     */
/************************************************************************/

EdgeList* 
AirScheduleSolu2::decideEdgeList(EdgeList* currELst, EdgeGraph *currEG, 
				 EdgeGraph *origEG, char * covered, int DEBUG)
{
    Edge* edg1;
    Edge* edg2;

    Pair pair;
    EdgeList *el;
    EdgeList *ret=NULL;
    int i, j, mxIndx, maxi;
    int num_of_ele = currELst->getNumberOfEdges();
    //tSet<Pair> pairLst(num_of_ele*num_of_ele);
    DcSortedSet<Pair,int> pairLst(num_of_ele*num_of_ele);
    float mxRotationTime = myProblem->getMaxRotationPeriod();

    if(DEBUG)
    {
       cout << " There are " << num_of_ele << " edges in " << endl;
       cout << *currELst << endl;

    }// end of if

    /*********************************************************************/
    /** Start from the first edge and find out all the sequences such as */
    /** it starts from the base city and ends at the same base city      */
    /*********************************************************************/

    for(i=0;i<num_of_ele;i++)
    {
       edg1 = currELst->getNthEdge(i);

       if(edg1->doesStartFromBaseCity())
       {
	  if(edg1->doesEndAtBaseCity() &&
	     edg1->getStartCityIndex() == edg1->getDestinationCityIndex() )
          {
	     Pair a(i,i);
	     //pairLst.addElement(a);
	     pairLst.addElement(a,1);

	     //if(DEBUG) cout << " from (" << i << " to " << i << ") is added!" << endl;

	  }// end of if

	  for(j=i+1;j<num_of_ele;j++)
	  {
	     edg2 = currELst->getNthEdge(j);

	     if(edg2->doesEndAtBaseCity() &&
	        edg1->getStartCityIndex() == edg2->getDestinationCityIndex() &&
		(edg2->getEndDate()-edg1->getStartDate())<=mxRotationTime)
	     {  
	        Pair a(i,j);
		//pairLst.addElement(a);
		pairLst.addElement(a,j-i+1);

		//if(DEBUG) cout << " from (" << i << " to " << j << ") is added!" ;
             } 

	  }// end of for

       }// end of if

    }// end of for

    /***************************************************/
    /** Restore Graph/covered for the deprecated edges */
    /***************************************************/

    putEdgesBackFromEdgeList(currELst,currEG,origEG,covered);

    /***************************/
    /** If there is a sequence */
    /***************************/
    if(pairLst.getSize()>0)
    {
	/*
        // *************************
	// ** Find the Longest One *
        // *************************
	mxIndx = 0;
	pairLst.getElementAtSafe(mxIndx,pair);
	maxi = pair.getTo()-pair.getFrom()+1;

	for(i=1;i<pairLst.getSize();i++)
	{
	   pairLst.getElementAtSafe(i,pair);

	   if((pair.getTo()-pair.getFrom()+1)>maxi)
	   {
	       maxi = pair.getTo()-pair.getFrom()+1;
	       mxIndx = i;

	   }// end of if

	}// end of for

        // *************************
	// ** get the longest pair *
        // *************************
	pairLst.getElementAtSafe(mxIndx,pair);

        // ****************************
	// ** Build the new edge list *
        // ****************************
	ret = new EdgeList(myProblem->getNumberOfFlights());
	for(i=pair.getFrom();i<=pair.getTo();i++)
  	        ret->addEdge(currELst->getNthEdge(i));

        // ****************************
	// ** Update the edge graph   *
        // ****************************
	if( ret->isValid(myProblem) ) putEdgeList(ret,currEG,covered);
	else { delete ret; ret = NULL; }
	*/

	int notfound = 1;
	ret = new EdgeList(myProblem->getNumberOfFlights());
	i=0;

	// Look at all the candidates starting from the maximum paths
	while( notfound && i<pairLst.getSize() )
	{
	   if( pairLst.getAt(i,pair) )
	   {
	      for(j=pair.getFrom();j<=pair.getTo();j++)
                  ret->addEdge(currELst->getNthEdge(j));

              if( ret->isValid(myProblem) )
	      {
		  putEdgeList(ret,currEG,covered);
		  notfound=0;
	      } 
	      else ret->clear();

	   }// end of if

	   i++;

	}// end of while

	// if there is nothing then reclaim the memory.
	if( notfound ) { delete ret; ret=NULL; }

    }// end of if

    return ret;

}// end of decideEdgeList()

/************************************************************************/
/*  */
/************************************************************************/
AirScheduleSolu2::findAllPaths(EdgeList* currELst, Pair & pair, int DEBUG)
{
   int ret = 0;

   Edge* edg1;
   Edge* edg2;
   int i, j, mxIndx, maxi;
   int num_of_ele = currELst->getNumberOfEdges();
   tSet<Pair> pairLst(num_of_ele*num_of_ele);

   if(DEBUG)
   {
       cout << " There are " << num_of_ele << " edges in " << endl;
       cout << *currELst << endl;

   }// end of if

   /*********************************************************************/
   /** Start from the first edge and find out all the sequences such as */
   /** it starts from the base city and ends at the same base city      */
   /*********************************************************************/

   for(i=0;i<num_of_ele;i++)
   {
       edg1 = currELst->getNthEdge(i);

       if(edg1->doesStartFromBaseCity())
       {
	  if(edg1->doesEndAtBaseCity() &&
	     edg1->getStartCityIndex() == edg1->getDestinationCityIndex() )
          {
	     Pair a(i,i);
	     pairLst.addElement(a);

	     //if(DEBUG) cout << " from (" << i << " to " << i << ") is added!" << endl;

	  }// end of if

	  for(j=i+1;j<num_of_ele;j++)
	  {
	     edg2 = currELst->getNthEdge(j);

	     if(edg2->doesEndAtBaseCity() &&
	        edg1->getStartCityIndex() == edg2->getDestinationCityIndex() )
	     {  
	        Pair a(i,j);
		pairLst.addElement(a);

		//if(DEBUG) cout << " from (" << i << " to " << j << ") is added!" ;
             } 

	  }// end of for

       }// end of if

   }// end of for

   /***************************/
   /** If there is a sequence */
   /***************************/
   if(pairLst.getSize()>0)
   {
	ret = 1;

        /*************************/
	/** Find the Longest One */
        /*************************/
	mxIndx = 0;
	pairLst.getElementAtSafe(mxIndx,pair);
	maxi = pair.getTo()-pair.getFrom()+1;

	for(i=1;i<pairLst.getSize();i++)
	{
	   pairLst.getElementAtSafe(i,pair);

	   if((pair.getTo()-pair.getFrom()+1)>maxi)
	   {
	       maxi = pair.getTo()-pair.getFrom()+1;
	       mxIndx = i;

	   }// end of if

	}// end of for

        /*************************/
	/** get the longest pair */
        /*************************/
	pairLst.getElementAtSafe(mxIndx,pair);

   }// end of if

   return ret;

}// end of findAllPaths()

/************************************************************************/
/*  */
/************************************************************************/
AirScheduleSolu2::findMaximumPath( EdgeList* currELst, Pair & pair, 
				   int STRICT, int DEBUG)
{
   int ret = 0;

   Edge* edg1;
   Edge* edg2;
   int i, j, mxIndx, maxi=0;
   int num_of_ele = currELst->getNumberOfEdges();
   float a;

   if(DEBUG)
   {
       cout << " There are " << num_of_ele << " edges in " << endl;
       cout << *currELst << endl;

   }// end of if

   /*********************************************************************/
   /** Start from the first edge and find out all the sequences such as */
   /** it starts from the base city and ends at the same base city      */
   /*********************************************************************/

   for(i=0;i<num_of_ele&&(i+maxi)<num_of_ele;i++)
   {
       edg1 = currELst->getNthEdge(i);

       if( STRICT || edg1->doesStartFromBaseCity() )
       {
	  if( ( STRICT || edg1->doesEndAtBaseCity() ) &&
	     edg1->getStartCityIndex() == edg1->getDestinationCityIndex() )
          {
	     Pair a(i,i);
	     pair = a;
	     ret = 1;
	     maxi = 1;

	     if(DEBUG)
                cout << " from (" << i << " to " << i << ") is maximum!" << endl;

	  }// end of if

	  //for(j=i+1;j<num_of_ele;j++)
	  for(j=num_of_ele-1;j>(i+maxi);j--)
	  {
	     edg2 = currELst->getNthEdge(j);

	     if( ( STRICT || edg2->doesEndAtBaseCity() ) &&
                 edg1->getStartCityIndex() == edg2->getDestinationCityIndex() )
	     {  
	        Pair a(i,j);
	        pair = a;
	        ret = 1;
	        maxi = j-i+1;

		//if(DEBUG) cout<<" from (" << i << " to " << j << ") is maximum!"<<endl ;
             } 

	  }// end of for

       }// end of if

   }// end of for

   return ret;

}// end of findMaximumPath()

/************************************************************************/
/*  */
/************************************************************************/

void
AirScheduleSolu2::binary(char *array, int from, int val, int maxdigit)
{
   int i = maxdigit-1; 
   int myval = val;
   int power;

   while(myval>0 && i>=0)
   {
     power = (int)pow((double)2,(double)i);
     if(myval>=power)
     {
	array[from+i] = 1;
	myval -= power;

     }// end of if

     i--;

   }// end of while

}// end of binary()

/************************************************************************/
/*  */
/************************************************************************/

void
AirScheduleSolu2::produceColumns()
{
   int numberOfFlights = myProblem->getNumberOfFlights();
   int i, j, k, m, n, stay, bits;
   char *covered = new char[numberOfFlights];
   Column* col;
   tSet<Column> ColSet;
   tSet<Column> Offsprings;
   Flight* base;
   Flight* f;

   for(i=0;i<numberOfFlights;i++) covered[i] = 0;

   for(i=0;i<numberOfFlights;i++)
   {
      covered[i] = 1;
       
      col = new Column(covered,numberOfFlights,myProblem);	


      ColSet.addElement(*col);
      cout << *col << " is added " << endl;

      j=i+1; stay=1;
      base = myProblem->getFlightWithIndex(i);
      while(stay && j<numberOfFlights)
      {
	 f = myProblem->getFlightWithIndex(j);

	 if((f->getArrivalDate()-base->getDepartureDate())<1*24*60) j++;
	 else stay=0;

      }// end of while

      cout << " for "<< i << " j = " << j << endl;
      bits = j - i +1;

      if(bits>0)
      {
	 m = (int) pow((double)2,(double)bits);
         cout << " m "<< m << endl;

	 for(k=1;k<m;k++)
	 {
            //cout << " k "<< k  << " for "<< i << " j = " << j << endl;
	    binary(covered,i+1,k,bits);
            //cout << " after binary "<< endl;

	    col = new Column(covered,numberOfFlights,myProblem);

	    if(col->isTrip()) 
	    {  ColSet.addElement(*col);
               cout << *col << " is added " << endl;
            }
	    else delete col;

	    for(n=i+1;n<j;n++) covered[n]=0;

	 }// end of for

      }// end of if

      covered[i] = 0;

   }// end of for

   Column *res;
   for(i=ColSet.getSize()-1;i>0;i--)
   {
     col = ColSet.getElementAt(i);
     for(j=i-1;j>=0;j--)
     {
        //res = col->doesCombine(*ColSet.getElementAt(j),myProblem);	
        res = col->doesCombine(*ColSet.getElementAt(j));	
	if(res!=NULL) Offsprings.addElement(*res);

     }// end of for

   }// end of for      

   ColSet = ColSet + Offsprings;
   Offsprings.clearAll();
    
   /** Populate them */

   Column *parent1;
   Column *parent2;
   char *coverParent1;
   char *coverParent2;
   char *coverOff1 = new char[numberOfFlights];
   char *coverOff2 = new char[numberOfFlights];
   int  pick1, pick2;
   int  nt0_1, nt0_2;

   cout << "Let's Populate them " << endl;

   for(i=0; i<5 && ColSet.getSize() < 2*numberOfFlights; i++)
   {
      for(j=0;j<numberOfFlights && 
	      (ColSet.getSize()+Offsprings.getSize()) < 2*numberOfFlights ;j++)
      {
	 pick2 = pick1 = rand() % ColSet.getSize();
	 while(pick2==pick1) pick2 = rand() % ColSet.getSize();

         cout << "j " << setw(3) << j << " pick1 ";
         cout << setw(3) << pick1 << " pick2 ";
         cout << setw(3) << pick2 << endl;

	 parent1 = ColSet.getElementAt(pick1);
	 parent2 = ColSet.getElementAt(pick2);
	 coverParent1 = parent1->getCover();
	 coverParent2 = parent2->getCover();
         cout << " I get the covers "<< endl;

	 pick1 = rand() % (numberOfFlights-2) + 1;
         cout << " cut point is " << pick1 << endl;
         nt0_1 = nt0_2 = 0;

	 for(k=0;k<pick1;k++)
	 {
	    coverOff1[k] = coverParent2[k];
	    coverOff2[k] = coverParent1[k];

	    if(coverOff1[k]>0) nt0_1 = 1;
	    if(coverOff2[k]>0) nt0_2 = 1;

	 }// end of for 

         cout << " cut point is " << pick1 ;
	 cout << " and numberOfFlights " << numberOfFlights << endl;

	 for(;k<numberOfFlights;k++)
	 {
	    coverOff1[k] = coverParent1[k];
	    coverOff2[k] = coverParent2[k];
	    if(coverOff1[k]>0) nt0_1 = 1;
	    if(coverOff2[k]>0) nt0_2 = 1;

	 }// end of for 

	 delete[] coverParent2;
	 delete[] coverParent1;

         cout << " after delete " << endl;

	 if(nt0_1>0)
	 {
	    res = new Column(coverOff1,numberOfFlights,myProblem);
            cout << " res1 " << *res << endl;
	    if( res->isTrip() && 
	     res->getMinNumMinsBtwFlights()<myProblem->getMinimumTime() &&
	     res->howManyMinutes() < 12 * 60 ) 
	     {
		cout << " A new Element " << *res << endl;
		Offsprings.addElement(*res);
             } 

            delete res; 

	 }// end of if

         cout << " after first child "<< endl;

	 if(nt0_2>0)
	 {
	    res = new Column(coverOff2,numberOfFlights,myProblem);
            cout << " res2 " << *res << endl;
	    if( res->isTrip() && 
	     res->getMinNumMinsBtwFlights()<myProblem->getMinimumTime() &&
	     res->howManyMinutes() < 12 * 60 ) 
	     {
		cout << " A new Element " << *res << endl;
		Offsprings.addElement(*res);
             } 

            delete res; 

	 }// end of if

         cout << " after second child "<< endl;

      }// end of for

      if(Offsprings.getSize()>0)
      {
         ColSet = ColSet + Offsprings;
         Offsprings.clearAll();

      }// end of if

      cout << setw(3) << i << " size = " << ColSet.getSize() << endl;

   }// end of for

   ofstream fout("column.txt"); 

   for(i=0;i<ColSet.getSize();i++)
   {
       //cout << setw(3) << i << " " << *ColSet.getElementAt(i) << endl;
       ColSet.getElementAt(i)->binaryPrint(fout);

   }// end of if

   for(i=0;i<Offsprings.getSize();i++)
   {
       //cout << setw(3) << i << " " << *Offsprings.getElementAt(i) << endl;
       Offsprings.getElementAt(i)->binaryPrint(fout);

   }// end of if

   fout.close(); 

}// end of produceColumns()

/************************************************************************/
/*
//
*/
/************************************************************************/

void 
AirScheduleSolu2::getScheduleFromColumnsFile(char* fileName) 
{  
   cost      = 0;
   calculated = 0;
   int index = 0;
   myEdgeLists = NULL;
   debug = 1;

   ifstream infile(fileName);

   int val, j,k;
   char *buf = new char[80];
   char line1[80];
   char line2[80];

   char ch;
   long int NumOfRows = myProblem->getNumberOfFlights();
   int NuOfVariables;
   long int vals[4];
   long int ColumnNum, NonZeros, Profit ;
   int RowIndex[50];
   int Exit = 0;
   ulong i;
   
   i=j=0;

   discardAllListOfEdgeLists();

   cout << "File name : " << fileName << endl;

   if(infile) cout << " File is Ok " << endl;

   buf = new char[NumOfRows+20];
   infile.getline(buf,NumOfRows+20);
   NuOfVariables = atoi(buf);
   cout << "NuOfVariables : " << NuOfVariables << endl;
   BitArray<char> **binCols = new BitArray<char>*[NuOfVariables];

   char *profit;
   i=0;

   EdgeList *el  = NULL;
   Edge *edge = NULL;
   int from,to;

   /** Read column information */
   while(!infile.eof()&&!Exit&&i<NuOfVariables)
   {  
      infile.getline(buf,NumOfRows+20);
      profit = buf+NumOfRows+1;
      buf[NumOfRows]=NULL;
      binCols[i] = new BitArray<char>(NumOfRows);

      for(j=0;j<NumOfRows;j++) if(buf[j]=='1') binCols[i]->set(j);

      i++;

   }// end of while

   infile.close();

   delete[] buf;

   /** Lets filter out the columns which has the same structure */

   char *cvr = new char[NumOfRows]; 
   for(i=0;i<NumOfRows;i++) cvr[i]=0;

   BitArray<char> _and(NumOfRows);
   i=0;
   while(i<NuOfVariables)
   {
      j=i+1;
      while(j<NuOfVariables)
      {
	 _and = *binCols[i] & *binCols[j];

	 if(_and==*binCols[i] || _and==*binCols[j] ) 
	 {
	     if(_and==*binCols[i])
	     {
	       delete binCols[i];
	       binCols[i] = binCols[j];
	     } 
	     else delete binCols[j];

	     for(k=j;k<(NuOfVariables-1);k++) binCols[k] = binCols[k+1];
	     NuOfVariables--;

	 }// end of if

	 j++;
	  
      }// end of while

      i++;
      
   }// end of while


   i=0;
   while(i<NuOfVariables)
   {  
      el = new EdgeList();
      from = to = -1;

      for(j=0;j<NumOfRows;j++) 
	 if(binCols[i]->test(j)) 
	 {
	    if(from==-1) { from = j; cout << endl << j ; }
	    else 
	     {
		if(edgeGraph!=NULL)
		{  
		   edge = new Edge(edgeGraph,from,j);
		   //cout << " Edge is " << *edge << endl;
		   if(edge!=NULL) el->addEdge(edge);
                }
		from = j;
	        cout << " -> " << j ;	

	     }

	 }// end of if

      if(el->getSize()) insertEdgeList(*el);

      cout << endl << buf << endl;
      cout << profit << endl;
      cout << *el << endl;

      i++;

   }// end of while

   /** free up the space */
   for(i=0;i<NuOfVariables;i++) delete binCols[i];
   delete[] binCols;

   delete[] cvr;

   ofstream ou1("redundancy.out");
   cout << " Before removeRedundancy() " << this << endl;
   ou1 << " Before removeRedundancy() " << this << endl;
   ou1.close();

   removeRedundancy();

   ofstream ou2("redundancy.out",ios::app);
   cout << " After removeRedundancy() " << this << endl;
   ou2 << " After removeRedundancy() " << this << endl;
   ou2.close();

}// end of getScheduleFromColumnsFile()

/************************************************************************/
/** Removes the Redundancies from the solution */
/************************************************************************/

void AirScheduleSolu2::removeRedundancy() 
{
   int ret = 0;
   int j, k, from, to, i=0,t;
   EdgeList* crrEdgeList;
   Edge*    edge;
   int  indx = getNumOfRotations();
   ListOfEdgeLists *tmp;
   int  removeIt;
   int  NumOfFlights = myProblem->getNumberOfFlights();
   int  numOfEdges;
   Edge*    tedge;
   char* covered = new char[NumOfFlights]; 
   char  myDEBUG = 0;

   for(i=0;i<NumOfFlights;i++) covered[i]=0;

   findCover(NumOfFlights,covered);

   ofstream ou1("redundancy.out"); //,ios::app);
   printCover(ou1,NumOfFlights,covered);

   /********************************/
   /** Find the List of edge lists */
   /********************************/
   for(i=0;i<getNumOfRotations();i++)
   {
      tmp = getListOfEdgeListsIndex(i);

      /*********************************/
      /** For each  List of Edge Lists */
      /*********************************/
      if( tmp!=NULL )
      {
      j=0;
      while( j<tmp->getNumberOfEdgeLists() )
      {
	 
         /***********************/
         /** For each Edge List */
         /***********************/
	 crrEdgeList = tmp->getNthEdgeList(j);
	 removeIt = 0;

	 if( crrEdgeList!=NULL )
	 {
	    removeIt = 1;

	    numOfEdges = crrEdgeList->getNumberOfEdges();

	    tedge = crrEdgeList->getFirstEdge();

	    if( numOfEdges==1 && 
		( covered[tedge->getStartNode()]<2 || 
		  covered[tedge->getEndNode()]<2 ) ) removeIt=0;
            else
	    {
	       if( covered[tedge->getStartNode()]<2 ) removeIt=0;

               for(t=0;t<numOfEdges-1&&removeIt;t++)
	       {
	          tedge = crrEdgeList->getNthEdge(t);
	          if( tedge!=NULL && covered[tedge->getEndNode()]<3 ) removeIt=0;

	       }// end of for
	       tedge = crrEdgeList->getNthEdge(t);
	       if( tedge!=NULL && removeIt && 
		   covered[tedge->getEndNode()]<2 ) removeIt=0;
	    }

         }// end of if

	 if( removeIt && crrEdgeList!=NULL )
	 {
	    if( myDEBUG )
	    { ou1 << " I can remove " << *crrEdgeList << endl
 		  << " from " << *tmp << endl;
	    }// end of if 

	    if( myDEBUG ) printCover(ou1,NumOfFlights,covered);

	    numOfEdges = crrEdgeList->getNumberOfEdges();

	    if( myDEBUG )
	    { ou1 << " There are " << numOfEdges << " edges " << endl;
	      ou1 << " I will remove " << *crrEdgeList << endl;
	    }// end of if 

	    tedge=NULL;

	    // Update Cover
	    for(t=0;t<numOfEdges;t++)
	    {
	       tedge = crrEdgeList->getNthEdge(t);
	       if( tedge!=NULL ) 
		   covered[tedge->getStartNode()] = ((int)covered[tedge->getStartNode()] - 1 );
	       if( myDEBUG )
	       {  ou1 << " covered[" << tedge->getStartNode() 
		      << "] : " << (int)covered[tedge->getStartNode()] << endl;
	       }// end of if 

	    }// end of for

	    if( tedge!=NULL ) 
	    {
	        covered[tedge->getEndNode()] = ((int)covered[tedge->getEndNode()] - 1 );
	        if( myDEBUG )
		{  ou1 << " covered[" << tedge->getEndNode() 
		       << "] : " << (int)covered[tedge->getEndNode()] << endl;
	        }// end of if 
            }

	    for(t=0;t<numOfEdges-1;t++)
	    {
	       tedge = crrEdgeList->getNthEdge(t);
	       if( tedge!=NULL ) 
		   covered[tedge->getEndNode()] = ((int)covered[tedge->getEndNode()] - 1 );
	       if( myDEBUG )
	       {  ou1 << " covered[" << tedge->getEndNode() 
		      << "] : " << (int)covered[tedge->getEndNode()] << endl;
	       }// end of if 

	    }// end of for

	    // Remove Element 
	    tmp->removeElementAt(j);

	    if( myDEBUG ) printCover(ou1,NumOfFlights,covered);
	    if( myDEBUG ) ou1 << " After removal " << endl << *tmp << endl;

	 } else j++;

      }// end of while

      }// end of if

   }// end of for

   if( myDEBUG ) cout << " Before Clean Up " << endl;

   i=0;
   while( i<getNumOfRotations() )
   {
      tmp = getListOfEdgeListsIndex(i);

      if( tmp!=NULL && tmp->getNumberOfEdgeLists()<1 ) 
	  myEdgeLists.removeElementAt(i);
      else i++;

   }// end of while

   ou1.close();

}// end of removeRedundancy()

/************************************************************************/
/** 
    Allows Dead-headings to cover the all assignments.
*/
/************************************************************************/

void AirScheduleSolu2::patchUpWithDeadHeadings(int NumOfFlights, char *covered)
{
   EdgeGraph currEG = *edgeGraph;
   int i, j;
   int rSeed = 0;
   
   ListOfEdgeLists cycles;
   ListOfEdgeLists paths;

   PathBuilder     pBuilder(myProblem);

   rSeed = RandomGenerator::generateInt(0,NumOfFlights);

   for( j=0;j<NumOfFlights;j++)
   {
     i = (j+rSeed)%NumOfFlights;

     if( covered[i]<1 )
     {
       cycles.reset();
       paths.reset();

       if( RandomGenerator::generateInt(0,1) )
       {

          //cout << " before buildPaths " << endl;
          pBuilder.buildPaths(currEG,cycles,paths,i,myProblem->getMaxRotationPeriod());
          //cout << " after buildPaths " << endl;

          if( cycles.getNumberOfEdgeLists()<1 )
          {
             //cout << " before buildPathsBackwards " << endl;
             pBuilder.buildPathsBackwards(currEG,cycles,paths,i,myProblem->getMaxRotationPeriod());
             //cout << " after buildPathsBackwards " << endl;

          }// end of if

       }
       else
	{

          //cout << " before buildPathsBackwards " << endl;
          pBuilder.buildPathsBackwards(currEG,cycles,paths,i,myProblem->getMaxRotationPeriod());
          //cout << " after buildPathsBackwards " << endl;

          if( cycles.getNumberOfEdgeLists()<1 )
          {
             //cout << " before buildPaths " << endl;
             pBuilder.buildPaths(currEG,cycles,paths,i,myProblem->getMaxRotationPeriod());
             //cout << " after buildPaths " << endl;

          }// end of if
	}

       if( cycles.getNumberOfEdgeLists()>0 )
       {
	  int peeki = 0;

	  if( cycles.getNumberOfEdgeLists()>1)
	      peeki = rand() % cycles.getNumberOfEdgeLists();

	  insertEdgeList(*cycles.getNthEdgeList(peeki));

	  if( calculatedCover )
	      myCover->updateCover(cycles.getNthEdgeList(peeki),
          			   NumOfFlights,covered);
	  else putEdgeList(cycles.getNthEdgeList(peeki),&currEG,covered);

       }
       else if( paths.getNumberOfEdgeLists()>0 )
	{
	   int peeki = 0;

	   if( paths.getNumberOfEdgeLists()>1) peeki = rand() % paths.getNumberOfEdgeLists();

	   insertEdgeList(*paths.getNthEdgeList(peeki));
	   if( calculatedCover )
	      myCover->updateCover( paths.getNthEdgeList(peeki), NumOfFlights,covered );
	   else putEdgeList(paths.getNthEdgeList(peeki),&currEG,covered);
	}

     }// end of if

   }// end of for

}// end of patchUpWithDeadHeadings()

/************************************************************************/
/** 
    Allows Dead-headings to cover the all assignments.
*/
/************************************************************************/

void AirScheduleSolu2::patchUpWithDeadHeadings1(int NumOfFlights, char *covered)
{
   EdgeGraph currEG = *edgeGraph;
   int i, j;
   int rSeed = 0;
   
   ListOfEdgeLists cycles;

   PathBuilder     pBuilder(myProblem);

   rSeed = RandomGenerator::generateInt(0,NumOfFlights);

   for( j=0;j<NumOfFlights;j++)
   {
     i = (j+rSeed)%NumOfFlights;

     if( covered[i]<1 )
     {
       cycles.reset();

       //cout << " before buildCycles " << endl;
       pBuilder.buildCycles(currEG,cycles,i,myProblem->getMaxRotationPeriod());
       //cout << " after buildCycles " << endl;

       if( cycles.getNumberOfEdgeLists()>0 )
       {
	  int peeki = 0;

	  if( cycles.getNumberOfEdgeLists()>1)
	      peeki = rand() % cycles.getNumberOfEdgeLists();

	  insertEdgeList(*cycles.getNthEdgeList(peeki));

	  if( calculatedCover )
	      myCover->updateCover(cycles.getNthEdgeList(peeki),
          			   NumOfFlights,covered);
	  else putEdgeList(cycles.getNthEdgeList(peeki),&currEG,covered);

       }// end of if

     }// end of if

   }// end of for

}// end of patchUpWithDeadHeadings1()


